# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward._dispatch import high_level_function
from awkward._layout import HighLevelContext, maybe_posaxis
from awkward._namedaxis import (
    _add_named_axis,
    _get_named_axis,
    _named_axis_to_positional_axis,
)
from awkward._nplikes.numpy_like import NumpyMetadata
from awkward._regularize import regularize_axis
from awkward.errors import AxisError

__all__ = ("singletons",)

np = NumpyMetadata.instance()


@high_level_function()
def singletons(array, axis=0, *, highlevel=True, behavior=None, attrs=None):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        axis (int or str): The dimension at which this operation is applied. The
            outermost dimension is `0`, followed by `1`, etc., and negative
            values count backward from the innermost: `-1` is the innermost
            dimension, `-2` is the next level up, etc.
            If a str, it is interpreted as the name of the axis which maps
            to an int if named axes are present. Named axes are attached
            to an array using #ak.with_named_axis and removed with
            #ak.without_named_axis; also see the
            [Named axes user guide](../../user-guide/how-to-array-properties-named-axis.html).
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        attrs (None or dict): Custom attributes for the output array, if
            high-level.

    Returns a singleton list (length 1) wrapping each non-missing value and
    an empty list (length 0) in place of each missing value.

    For example,

        >>> array = ak.Array([1.1, 2.2, None, 3.3, None, None, 4.4, 5.5])
        >>> ak.singletons(array).show()
        [[1.1],
         [2.2],
         [],
         [3.3],
         [],
         [],
         [4.4],
         [5.5]]

    See #ak.firsts to invert this function.
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, axis, highlevel, behavior, attrs)


def _impl(array, axis, highlevel, behavior, attrs):
    with HighLevelContext(behavior=behavior, attrs=attrs) as ctx:
        layout = ctx.unwrap(array, allow_record=False, primitive_policy="error")

    # Handle named axis
    named_axis = _get_named_axis(ctx)
    # Step 1: Normalize named axis to positional axis
    axis = _named_axis_to_positional_axis(named_axis, axis)

    axis = regularize_axis(axis, none_allowed=False)

    # Step 2: propagate named axis from input to output,
    #   use strategy "add one" (see: awkward._namedaxis)
    out_named_axis = _add_named_axis(
        named_axis, (axis + 1) if axis >= 0 else axis, layout.minmax_depth[1]
    )

    def action(layout, depth, backend, **kwargs):
        posaxis = maybe_posaxis(layout, axis, depth)

        if posaxis is not None and posaxis + 1 == depth:
            if layout.is_union or layout.is_record:
                return None

            elif layout.is_option:
                offsets = backend.nplike.empty(layout.length + 1, dtype=np.int64)
                offsets[0] = 0

                backend.nplike.cumsum(
                    layout.mask_as_bool(valid_when=True), maybe_out=offsets[1:]
                )

                return ak.contents.ListOffsetArray(
                    ak.index.Index64(offsets), layout.project()
                )

            else:
                return ak.contents.RegularArray(layout, 1).to_ListOffsetArray64(True)

        elif layout.is_leaf:
            raise AxisError(f"axis={axis} exceeds the depth of this array ({depth})")

    out = ak._do.recursively_apply(layout, action, numpy_to_regular=True)

    wrapped_out = ctx.wrap(
        out,
        highlevel=highlevel,
    )

    # propagate named axis to output
    return ak.operations.ak_with_named_axis._impl(
        wrapped_out,
        named_axis=out_named_axis,
        highlevel=highlevel,
        behavior=ctx.behavior,
        attrs=ctx.attrs,
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import pytest

import awkward as ak


def test_ravel_boolean():
    x = ak.Array([None, [False], [True]])
    y = ak.ravel(ak.fill_none(x, False, axis=0))
    assert y.tolist() == [False, False, True]


def test_ravel_boolean_typetracer():
    x = ak.Array([None, [False], [True]], backend="typetracer")
    y = ak.ravel(ak.fill_none(x, False, axis=0))
    assert y.typestr == "## * bool"


def test_ravel_ListOffsetArray():
    a = ak.Array([[1, 2, 3], 4, [5, 6], 7, 8])
    b = ak.ravel(a)
    assert b.tolist() == [1, 2, 3, 4, 5, 6, 7, 8]


def test_ravel_ListOffsetArray_typetracer():
    a = ak.Array([[1, 2, 3], 4, [5, 6], 7, 8], backend="typetracer")
    b = ak.ravel(a)
    assert b.typestr == "## * int64"


def test_ravel_nested_ListOffsetArrays():
    a = ak.Array([[[1, 2], 3], 4, [5, [[[6]]]], 7, 8])
    b = ak.ravel(a)
    assert b.tolist() == [1, 2, 3, 4, 5, 6, 7, 8]


def test_ravel_nested_ListOffsetArrays_typetracer():
    a = ak.Array([[[1, 2], 3], 4, [5, [[[6]]]], 7, 8], backend="typetracer")
    b = ak.ravel(a)
    assert b.typestr == "## * int64"


def test_ravel_incompatible_contents():
    a = ak.Array([1, "high", 2, "low"])
    with pytest.raises(
        AssertionError,
        match="cannot merge NumpyArray with ListOffsetArray",
    ):
        ak.ravel(a)


def test_ravel_incompatible_contents_typetracer():
    a = ak.Array([1, "high", 2, "low"], backend="typetracer")
    with pytest.raises(
        AssertionError,
        match="cannot merge NumpyArray with ListOffsetArray",
    ):
        ak.ravel(a)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy
import pytest

import awkward as ak
from awkward._nplikes.shape import unknown_length
from awkward._nplikes.typetracer import TypeTracerArray

# tests taken from tests/test_2365_enforce_type.py and adapted for typetracer arrays


@pytest.mark.parametrize("forget_length", [False, True])
def test_record(forget_length):
    ## record → record
    original = ak.to_layout([{"x": [1, 2]}], regulararray=False)
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(
        array, ak.types.from_datashape("{x: var * int64}", highlevel=False)
    )
    expected = ak.contents.RecordArray(
        [
            ak.contents.ListOffsetArray(
                ak.index.Index(numpy.array([0, 2], dtype=numpy.int64)),
                ak.contents.NumpyArray(numpy.array([1, 2], dtype=numpy.int64)),
            )
        ],
        ["x"],
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    ## record → record
    original = ak.to_layout([{"x": [1, 0]}], regulararray=False)
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(
        array, ak.types.from_datashape("{x: var * bool}", highlevel=False)
    )
    expected = ak.contents.RecordArray(
        [
            ak.contents.ListOffsetArray(
                ak.index.Index(numpy.array([0, 2], dtype=numpy.int64)),
                ak.contents.NumpyArray(numpy.array([1, 0], dtype=numpy.bool_)),
            )
        ],
        ["x"],
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    ## record → different tuple
    original = ak.to_layout([{"x": [1, 2]}], regulararray=False)
    array = original.to_typetracer(forget_length)
    with pytest.raises(ValueError, match=r"converted between records and tuples"):
        ak.enforce_type(
            array, ak.types.from_datashape("(var * float64)", highlevel=False)
        )

    original = ak.to_layout([{"x": [1, 2]}], regulararray=False)
    array = original.to_typetracer(forget_length)
    with pytest.raises(
        TypeError, match=r"can only add new fields to a record if they are option types"
    ):
        ak.enforce_type(
            array, ak.types.from_datashape("{y: var * float64}", highlevel=False)
        )

    ## record → totally different record
    original = ak.to_layout([{"x": [1, 2]}], regulararray=False)
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(
        array, ak.types.from_datashape("{y: ?var * float64}", highlevel=False)
    )
    expected = ak.contents.RecordArray(
        [
            ak.contents.IndexedOptionArray(
                ak.index.Index64([-1]),
                ak.contents.ListOffsetArray(
                    ak.index.Index(numpy.array([0], dtype=numpy.int64)),
                    ak.contents.NumpyArray(numpy.array([], dtype=numpy.float64)),
                ),
            )
        ],
        ["y"],
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    ## record → extended record
    original = ak.to_layout([{"x": [1, 2]}], regulararray=False)
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(
        array,
        ak.types.from_datashape("{x: var * int64, y: ?int64}", highlevel=False),
    )
    expected = ak.contents.RecordArray(
        [
            ak.contents.ListOffsetArray(
                ak.index.Index(numpy.array([0, 2], dtype=numpy.int64)),
                ak.contents.NumpyArray(numpy.array([1, 2], dtype=numpy.int64)),
            ),
            ak.contents.IndexedOptionArray(
                ak.index.Index64([-1]),
                ak.contents.NumpyArray(numpy.array([], dtype=numpy.int64)),
            ),
        ],
        ["x", "y"],
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    ## record → empty record
    original = ak.to_layout([{"x": [1, 2]}], regulararray=False)
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(array, ak.types.from_datashape("{}", highlevel=False))
    expected = ak.contents.RecordArray([], [], length=1).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    ############

    ## tuple → tuple
    original = ak.to_layout([([1, 2],)], regulararray=False)
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(
        array, ak.types.from_datashape("(var * int64)", highlevel=False)
    )
    expected = ak.contents.RecordArray(
        [
            ak.contents.ListOffsetArray(
                ak.index.Index(numpy.array([0, 2], dtype=numpy.int64)),
                ak.contents.NumpyArray(numpy.array([1, 2], dtype=numpy.int64)),
            )
        ],
        None,
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    ## tuple → tuple
    original = ak.to_layout([([1, 0],)], regulararray=False)
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(
        array, ak.types.from_datashape("(var * bool)", highlevel=False)
    )
    expected = ak.contents.RecordArray(
        [
            ak.contents.ListOffsetArray(
                ak.index.Index(numpy.array([0, 2], dtype=numpy.int64)),
                ak.contents.NumpyArray(numpy.array([1, 0], dtype=numpy.bool_)),
            )
        ],
        None,
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    ## tuple → different record
    original = ak.to_layout([([1, 2],)], regulararray=False)
    array = original.to_typetracer(forget_length)
    with pytest.raises(ValueError, match=r"converted between records and tuples"):
        ak.enforce_type(
            array, ak.types.from_datashape("{x: var * float64}", highlevel=False)
        )

    original = ak.to_layout([([1, 2],)], regulararray=False)
    array = original.to_typetracer(forget_length)
    with pytest.raises(
        TypeError, match=r"can only add new slots to a tuple if they are option types"
    ):
        ak.enforce_type(
            array, ak.types.from_datashape("(var * int64, float32)", highlevel=False)
        )

    ## tuple → extended tuple
    original = ak.to_layout([([1, 2],)], regulararray=False)
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(
        array, ak.types.from_datashape("(var * int64, ?float32)", highlevel=False)
    )
    expected = ak.contents.RecordArray(
        [
            ak.contents.ListOffsetArray(
                ak.index.Index(numpy.array([0, 2], dtype=numpy.int64)),
                ak.contents.NumpyArray(numpy.array([1, 2], dtype=numpy.int64)),
            ),
            ak.contents.IndexedOptionArray(
                ak.index.Index64([-1]),
                ak.contents.NumpyArray(numpy.array([], dtype=numpy.float32)),
            ),
        ],
        None,
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    ## tuple → empty tuple
    original = ak.to_layout([([1, 2],)], regulararray=False)
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(array, ak.types.from_datashape("()", highlevel=False))
    expected = ak.contents.RecordArray([], None, length=1).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)


@pytest.mark.parametrize("forget_length", [False, True])
def test_list(forget_length):
    original = ak.to_layout([[1, 2, 3]])
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(
        array, ak.types.from_datashape("var * int64", highlevel=False)
    )
    expected = ak.contents.ListOffsetArray(
        ak.index.Index(numpy.array([0, 3], dtype=numpy.int64)),
        ak.contents.NumpyArray(numpy.array([1, 2, 3], dtype=numpy.int64)),
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    original = ak.to_layout([[1, 2, 3]])
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(
        array, ak.types.from_datashape("3 * int64", highlevel=False)
    )
    expected = ak.contents.RegularArray(
        ak.contents.NumpyArray(numpy.array([1, 2, 3], dtype=numpy.int64)), size=3
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    ## Empty list to regular shape
    original = ak.to_layout([[]])[:0]
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(
        array, ak.types.from_datashape("3 * int64", highlevel=False)
    )
    expected = ak.contents.RegularArray(
        ak.contents.NumpyArray(numpy.array([], dtype=numpy.int64)), size=3
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    original = ak.to_layout([[1, 2, 3]])
    array = original.to_typetracer(forget_length)
    expected = ak.contents.RegularArray(
        ak.contents.NumpyArray(
            TypeTracerArray._new(numpy.dtype("int64"), (unknown_length,))
        ),
        size=unknown_length,
    )
    assert result.layout.is_equal_to(expected)

    original = ak.to_regular([[1, 2, 3]], axis=-1, highlevel=False)
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(
        array, ak.types.from_datashape("var * int64", highlevel=False)
    )
    expected = ak.contents.ListOffsetArray(
        ak.index.Index(numpy.array([0, 3], dtype=numpy.int64)),
        ak.contents.NumpyArray(numpy.array([1, 2, 3], dtype=numpy.int64)),
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    original = ak.to_regular([[1, 2, 3]], axis=-1, highlevel=False)
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(
        array, ak.types.from_datashape("3 * int64", highlevel=False)
    )
    expected = ak.contents.RegularArray(
        ak.contents.NumpyArray(numpy.array([1, 2, 3], dtype=numpy.int64)), size=3
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    original = ak.to_regular([[1, 2, 3]], axis=-1, highlevel=False)
    array = original.to_typetracer(forget_length)
    with pytest.raises(ValueError, match=r"different size"):
        ak.enforce_type(array, ak.types.from_datashape("4 * int64", highlevel=False))


@pytest.mark.parametrize("forget_length", [False, True])
def test_option(forget_length):
    ## option → option
    original = ak.to_layout([1, None])
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(array, ak.types.from_datashape("?int64", highlevel=False))
    expected = ak.contents.IndexedOptionArray(
        ak.index.Index(numpy.array([0, -1], dtype=numpy.int64)),
        ak.contents.NumpyArray(numpy.array([1], dtype=numpy.int64)),
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    ## option → option (packing)
    original = ak.to_layout([1, None, 2, 3])
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(
        array, ak.types.from_datashape("?float64", highlevel=False)
    )
    expected = ak.contents.IndexedOptionArray(
        ak.index.Index(numpy.array([0, -1, 1, 2], dtype=numpy.int64)),
        ak.contents.NumpyArray(numpy.array([1, 2, 3], dtype=numpy.float64)),
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    ## option → no option
    original = ak.to_layout([1, None])[:1]
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(array, ak.types.from_datashape("int64", highlevel=False))
    expected = ak.contents.IndexedArray(
        ak.index.Index(numpy.array([0], dtype=numpy.int64)),
        ak.contents.NumpyArray(numpy.array([1], dtype=numpy.int64)),
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    original = ak.to_layout([1, None])
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(array, ak.types.from_datashape("int64", highlevel=False))
    expected = ak.contents.IndexedArray(
        ak.index.Index(numpy.array([0], dtype=numpy.int64)),
        ak.contents.NumpyArray(numpy.array([1], dtype=numpy.int64)),
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    ## Add option
    original = ak.to_layout([1, 2])
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(array, ak.types.from_datashape("?int64", highlevel=False))
    expected = ak.contents.UnmaskedArray(
        ak.contents.NumpyArray(numpy.array([1, 2], dtype=numpy.int64))
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    ## option[X] → option[unknown]
    original = ak.to_layout([None, 1, 2, 3])
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(array, "?unknown")
    expected = ak.contents.IndexedOptionArray(
        ak.index.Index64([-1, -1, -1, -1]), ak.contents.EmptyArray()
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)


@pytest.mark.parametrize("forget_length", [False, True])
def test_numpy(forget_length):
    ## NumPy
    ## 1D → 1D
    original = ak.to_layout([1, 2])
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(array, ak.types.from_datashape("int64", highlevel=False))
    expected = ak.contents.NumpyArray(
        numpy.array([1, 2], dtype=numpy.int64)
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    original = ak.to_layout([1, 2])
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(array, ak.types.from_datashape("float32", highlevel=False))
    expected = ak.contents.NumpyArray(
        numpy.array([1.0, 2.0], dtype=numpy.float32)
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    original = ak.to_layout([1, 2])
    array = original.to_typetracer(forget_length)
    with pytest.raises(TypeError):
        ak.enforce_type(array, ak.types.from_datashape("string", highlevel=False))

    ## 1D → 2D
    original = ak.to_layout([1, 2])
    array = original.to_typetracer(forget_length)
    with pytest.raises(TypeError):
        ak.enforce_type(array, ak.types.from_datashape("var * int64", highlevel=False))

    original = ak.to_layout([1, 2])
    array = original.to_typetracer(forget_length)
    with pytest.raises(TypeError):
        ak.enforce_type(array, ak.types.from_datashape("2 * float32", highlevel=False))

    ## 2D → 1D
    original = ak.to_layout(numpy.zeros((2, 3)), regulararray=False)
    array = original.to_typetracer(forget_length)
    with pytest.raises(TypeError):
        ak.enforce_type(array, ak.types.from_datashape("int64", highlevel=False))

    original = ak.to_layout(numpy.zeros((2, 3)), regulararray=False)
    array = original.to_typetracer(forget_length)
    with pytest.raises(TypeError):
        ak.enforce_type(array, ak.types.from_datashape("float32", highlevel=False))

    ## 2D → 2D
    original = ak.to_layout(numpy.zeros((2, 3)), regulararray=False)
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(
        array, ak.types.from_datashape("var * int64", highlevel=False)
    )
    expected = ak.contents.ListOffsetArray(
        ak.index.Index(numpy.array([0, 3, 6], dtype=numpy.int64)),
        ak.contents.NumpyArray(numpy.array([0, 0, 0, 0, 0, 0], dtype=numpy.int64)),
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    original = ak.to_layout(numpy.zeros((2, 3)), regulararray=False)
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(
        array, ak.types.from_datashape("3 * float32", highlevel=False)
    )
    expected = ak.contents.RegularArray(
        ak.contents.NumpyArray(numpy.array([0, 0, 0, 0, 0, 0], dtype=numpy.float32)),
        size=3,
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)


@pytest.mark.parametrize("forget_length", [False, True])
def test_union(forget_length):
    ## non union → union
    original = ak.to_layout([1, 2])
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(
        array, ak.types.from_datashape("union[int64, string]", highlevel=False)
    )
    expected = ak.contents.UnionArray(
        tags=ak.index.Index8([0, 0]),
        index=ak.index.Index64([0, 1]),
        contents=[
            ak.contents.NumpyArray(numpy.array([1, 2], dtype=numpy.int64)),
            ak.contents.ListOffsetArray(
                offsets=ak.index.Index64([0]),
                content=ak.contents.NumpyArray(
                    numpy.array([], dtype=numpy.uint8),
                    parameters={"__array__": "char"},
                ),
                parameters={"__array__": "string"},
            ),
        ],
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    original = ak.to_layout([1, 2])
    array = original.to_typetracer(forget_length)
    with pytest.raises(TypeError):
        ak.enforce_type(
            array,
            ak.types.from_datashape("union[var * int64, string]", highlevel=False),
        )

    ## union → no union (project)
    original = ak.to_layout([1, "hi", "bye"])[1:2]
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(array, ak.types.from_datashape("string", highlevel=False))
    expected = ak.contents.IndexedArray(
        ak.index.Index64([0]),
        ak.contents.ListOffsetArray(
            offsets=ak.index.Index64([0, 2, 5]),
            content=ak.contents.NumpyArray(
                numpy.array([104, 105, 98, 121, 101], dtype=numpy.uint8),
                parameters={"__array__": "char"},
            ),
            parameters={"__array__": "string"},
        ),
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    original = ak.to_layout([1, "hi", "bye"])[:1]
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(array, ak.types.from_datashape("int64", highlevel=False))
    expected = ak.contents.IndexedArray(
        ak.index.Index64([0]),
        ak.contents.NumpyArray(numpy.array([1], dtype=numpy.int64)),
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    original = ak.to_layout([1, "hi"])
    array = original.to_typetracer(forget_length)
    with pytest.raises(TypeError):
        ak.enforce_type(array, ak.types.from_datashape("var * int64", highlevel=False))

    ## union → no union (convert)
    original = ak.concatenate(
        [
            [{"x": 1, "y": 2.0}],
            [{"y": 3}],
            [{"x": 4, "y": 5.0}],
            [{"y": 6}],
        ],
    )
    array = original.layout.to_typetracer(forget_length)
    result = ak.enforce_type(
        array, ak.types.from_datashape("{y: int64}", highlevel=False)
    )
    expected = ak.contents.IndexedArray(
        index=ak.index.Index64([0, 2, 1, 3]),
        content=ak.contents.RecordArray(
            contents=[
                ak.contents.NumpyArray(
                    numpy.array([2, 5, 3, 6], dtype=numpy.int64),
                )
            ],
            fields=["y"],
        ),
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    original = ak.to_layout([1, "hi"])
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(array, ak.types.from_datashape("string", highlevel=False))
    expected = ak.contents.IndexedArray(
        ak.index.Index64([0]),
        ak.contents.ListOffsetArray(
            offsets=ak.index.Index64([0, 2]),
            content=ak.contents.NumpyArray(
                numpy.array([104, 105], dtype=numpy.uint8),
                parameters={"__array__": "char"},
            ),
            parameters={"__array__": "string"},
        ),
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    ## union → same union
    original = ak.to_layout([1, "hi"])
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(
        array,
        ak.types.UnionType(
            [
                ak.types.NumpyType("int64"),
                ak.types.ListType(
                    ak.types.NumpyType("uint8", parameters={"__array__": "char"}),
                    parameters={"__array__": "string", "foo": "bar"},
                ),
            ]
        ),
    )
    expected = ak.contents.UnionArray(
        tags=ak.index.Index8([0, 1]),
        index=ak.index.Index64([0, 0]),
        contents=[
            ak.contents.NumpyArray(numpy.array([1], dtype=numpy.int64)),
            ak.contents.ListOffsetArray(
                offsets=ak.index.Index64([0, 2]),
                content=ak.contents.NumpyArray(
                    numpy.array([104, 105], dtype=numpy.uint8),
                    parameters={"__array__": "char"},
                ),
                parameters={"__array__": "string", "foo": "bar"},
            ),
        ],
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    ## union → bigger union
    original = ak.to_layout([1, "hi"])
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(
        array,
        ak.types.from_datashape("union[int64, string, datetime64]", highlevel=False),
    )
    expected = ak.contents.UnionArray(
        tags=ak.index.Index8([0, 1]),
        index=ak.index.Index64([0, 0]),
        contents=[
            ak.contents.NumpyArray(numpy.array([1], dtype=numpy.int64)),
            ak.contents.ListOffsetArray(
                offsets=ak.index.Index64([0, 2]),
                content=ak.contents.NumpyArray(
                    numpy.array([104, 105], dtype=numpy.uint8),
                    parameters={"__array__": "char"},
                ),
                parameters={"__array__": "string"},
            ),
            ak.contents.NumpyArray(numpy.array([], dtype=numpy.datetime64)),
        ],
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    ## union → different union (same N)
    original = ak.to_layout([1, "hi"])
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(
        array,
        ak.types.UnionType(
            [
                ak.types.NumpyType("float32"),
                ak.types.ListType(
                    ak.types.NumpyType("uint8", parameters={"__array__": "char"}),
                    parameters={"__array__": "string", "foo": "bar"},
                ),
            ]
        ),
    )
    expected = ak.contents.UnionArray(
        tags=ak.index.Index8([0, 1]),
        index=ak.index.Index64([0, 0]),
        contents=[
            ak.contents.NumpyArray(numpy.array([1], dtype=numpy.float32)),
            ak.contents.ListOffsetArray(
                offsets=ak.index.Index64([0, 2]),
                content=ak.contents.NumpyArray(
                    numpy.array([104, 105], dtype=numpy.uint8),
                    parameters={"__array__": "char"},
                ),
                parameters={"__array__": "string", "foo": "bar"},
            ),
        ],
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    ## union → different union (smaller N)
    original = ak.to_layout([1, "hi", [1j, 2j]])[:2]
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(array, "union[int64, string]")
    expected = ak.contents.UnionArray(
        tags=ak.index.Index8([0, 1]),
        index=ak.index.Index64([0, 0]),
        contents=[
            ak.contents.NumpyArray(numpy.array([1], dtype=numpy.int64)),
            ak.contents.ListOffsetArray(
                offsets=ak.index.Index64([0, 2]),
                content=ak.contents.NumpyArray(
                    numpy.array([104, 105], dtype=numpy.uint8),
                    parameters={"__array__": "char"},
                ),
                parameters={"__array__": "string"},
            ),
        ],
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    ## union → incompatible different union (same N)
    original = ak.to_layout([1, "hi"])
    array = original.to_typetracer(forget_length)
    with pytest.raises(TypeError):
        ak.enforce_type(
            array, ak.types.from_datashape("union[int64, bool]", highlevel=False)
        )

    ## union → different union (same N, more than one change)
    original = ak.to_layout([1, "hi", False])
    array = original.to_typetracer(forget_length)
    with pytest.raises(TypeError):
        ak.enforce_type(
            array,
            ak.types.from_datashape(
                "union[datetime64, string, float32]", highlevel=False
            ),
        )

    ## union of union → union of extended union
    original = ak.contents.UnionArray(
        ak.index.Index8([0, 1]),
        ak.index.Index64([0, 0]),
        [
            ak.contents.NumpyArray(numpy.array([1], dtype=numpy.int64)),
            ak.contents.ListOffsetArray(
                ak.index.Index64([0, 2]),
                ak.contents.UnionArray(
                    ak.index.Index8([0, 1]),
                    ak.index.Index64([0, 0]),
                    [
                        ak.contents.NumpyArray(numpy.array([2], dtype=numpy.int64)),
                        ak.contents.ListOffsetArray(
                            ak.index.Index64([0, 2]),
                            ak.contents.RecordArray(
                                [
                                    ak.contents.IndexedOptionArray(
                                        ak.index.Index64([0, -1]),
                                        ak.contents.NumpyArray(
                                            numpy.array([1], dtype=numpy.int64)
                                        ),
                                    ),
                                    ak.contents.IndexedOptionArray(
                                        ak.index.Index64([-1, 0]),
                                        ak.contents.NumpyArray(
                                            numpy.array([2], dtype=numpy.int64)
                                        ),
                                    ),
                                ],
                                ["x", "y"],
                            ),
                        ),
                    ],
                ),
            ),
        ],
    )
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(
        array,
        """
    union[
        int64,
        var * union[
            int64,
            var * {
                x: ?int64,
                y: ?int64,
                z: ?string
            }
        ]
    ]
                             """,
    )
    expected = ak.contents.UnionArray(
        ak.index.Index8([0, 1]),
        ak.index.Index64([0, 0]),
        [
            ak.contents.NumpyArray(numpy.array([1], dtype=numpy.int64)),
            ak.contents.ListOffsetArray(
                ak.index.Index64([0, 2]),
                ak.contents.UnionArray(
                    ak.index.Index8([0, 1]),
                    ak.index.Index64([0, 0]),
                    [
                        ak.contents.NumpyArray(numpy.array([2], dtype=numpy.int64)),
                        ak.contents.ListOffsetArray(
                            ak.index.Index64([0, 2]),
                            ak.contents.RecordArray(
                                [
                                    ak.contents.IndexedOptionArray(
                                        ak.index.Index64([0, -1]),
                                        ak.contents.NumpyArray(
                                            numpy.array([1], dtype=numpy.int64)
                                        ),
                                    ),
                                    ak.contents.IndexedOptionArray(
                                        ak.index.Index64([-1, 0]),
                                        ak.contents.NumpyArray(
                                            numpy.array([2], dtype=numpy.int64)
                                        ),
                                    ),
                                    ak.contents.IndexedOptionArray(
                                        ak.index.Index64([-1, -1]),
                                        ak.contents.ListOffsetArray(
                                            ak.index.Index64([0]),
                                            ak.contents.NumpyArray(
                                                numpy.empty(0, dtype=numpy.uint8),
                                                parameters={"__array__": "char"},
                                            ),
                                            parameters={"__array__": "string"},
                                        ),
                                    ),
                                ],
                                ["x", "y", "z"],
                            ),
                        ),
                    ],
                ),
            ),
        ],
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)


@pytest.mark.parametrize("forget_length", [False, True])
def test_string(forget_length):
    ## string -> bytestring
    original = ak.to_layout(["hello world"])
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(array, ak.types.from_datashape("bytes", highlevel=False))
    expected = ak.contents.ListOffsetArray(
        offsets=ak.index.Index64([0, 11]),
        content=ak.contents.NumpyArray(
            numpy.array(
                [104, 101, 108, 108, 111, 32, 119, 111, 114, 108, 100],
                dtype=numpy.uint8,
            ),
            parameters={"__array__": "byte"},
        ),
        parameters={"__array__": "bytestring"},
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    ## bytestring -> string
    original = ak.to_layout([b"hello world"])
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(array, ak.types.from_datashape("string", highlevel=False))
    expected = ak.contents.ListOffsetArray(
        offsets=ak.index.Index64([0, 11]),
        content=ak.contents.NumpyArray(
            numpy.array(
                [104, 101, 108, 108, 111, 32, 119, 111, 114, 108, 100],
                dtype=numpy.uint8,
            ),
            parameters={"__array__": "char"},
        ),
        parameters={"__array__": "string"},
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    ## string -> string
    original = ak.to_layout(["hello world"])
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(array, ak.types.from_datashape("string", highlevel=False))
    expected = ak.contents.ListOffsetArray(
        offsets=ak.index.Index64([0, 11]),
        content=ak.contents.NumpyArray(
            numpy.array(
                [104, 101, 108, 108, 111, 32, 119, 111, 114, 108, 100],
                dtype=numpy.uint8,
            ),
            parameters={"__array__": "char"},
        ),
        parameters={"__array__": "string"},
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    ## bytestring -> bytestring
    original = ak.to_layout([b"hello world"])
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(array, ak.types.from_datashape("bytes", highlevel=False))
    expected = ak.contents.ListOffsetArray(
        offsets=ak.index.Index64([0, 11]),
        content=ak.contents.NumpyArray(
            numpy.array(
                [104, 101, 108, 108, 111, 32, 119, 111, 114, 108, 100],
                dtype=numpy.uint8,
            ),
            parameters={"__array__": "byte"},
        ),
        parameters={"__array__": "bytestring"},
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    ## bytestring -> list of byte
    original = ak.to_layout([b"hello world"])
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(
        array, ak.types.from_datashape("var * byte", highlevel=False)
    )
    expected = ak.contents.ListOffsetArray(
        offsets=ak.index.Index64([0, 11]),
        content=ak.contents.NumpyArray(
            numpy.array(
                [104, 101, 108, 108, 111, 32, 119, 111, 114, 108, 100],
                dtype=numpy.uint8,
            ),
            parameters={"__array__": "byte"},
        ),
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    ## bytestring -> list of int64
    original = ak.to_layout([b"hello world"])
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(
        array, ak.types.from_datashape("var * int64", highlevel=False)
    )
    expected = ak.contents.ListOffsetArray(
        offsets=ak.index.Index64([0, 11]),
        content=ak.contents.NumpyArray(
            numpy.array(
                [104, 101, 108, 108, 111, 32, 119, 111, 114, 108, 100],
                dtype=numpy.int64,
            )
        ),
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    ## list of int64 -> string
    original = ak.without_parameters([b"hello world"])
    array = original.layout.to_typetracer(forget_length)
    result = ak.enforce_type(array, ak.types.from_datashape("string", highlevel=False))
    expected = ak.contents.ListOffsetArray(
        offsets=ak.index.Index64([0, 11]),
        content=ak.contents.NumpyArray(
            numpy.array(
                [104, 101, 108, 108, 111, 32, 119, 111, 114, 108, 100],
                dtype=numpy.uint8,
            ),
            parameters={"__array__": "char"},
        ),
        parameters={"__array__": "string"},
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)


@pytest.mark.parametrize("forget_length", [False, True])
def test_highlevel(forget_length):
    original = ak.to_layout(["hello world"])
    array = original.to_typetracer(forget_length)
    with pytest.raises(TypeError, match=r"High-level type objects are not supported"):
        ak.enforce_type(array, ak.types.from_datashape("1 * bytes"))

    original = ak.to_layout([{"msg": "hello world"}])
    array = original.to_typetracer(forget_length)
    with pytest.raises(TypeError, match=r"High-level type objects are not supported"):
        ak.enforce_type(
            array,
            ak.types.ScalarType(
                ak.types.RecordType(
                    [
                        ak.types.ListType(
                            ak.types.NumpyType(
                                "uint8", parameters={"__array__": "char"}
                            ),
                            parameters={"__array__": "string"},
                        )
                    ],
                    ["msg"],
                )
            ),
        )


@pytest.mark.parametrize("forget_length", [False, True])
def test_single_record(forget_length):
    original = ak.to_regular([{"x": [1, 2]}])
    record = original.layout.to_typetracer(forget_length)[0]
    result = ak.enforce_type(
        record,
        ak.types.from_datashape("{x: var * float64}", highlevel=False),
        highlevel=False,
    )
    expected = ak.contents.RecordArray(
        [
            ak.contents.ListOffsetArray(
                ak.index.Index(numpy.array([0, 2], dtype=numpy.int64)),
                ak.contents.NumpyArray(numpy.array([1, 2], dtype=numpy.float64)),
            )
        ],
        ["x"],
    ).to_typetracer(forget_length)
    assert result.array.is_equal_to(expected)


@pytest.mark.parametrize("forget_length", [False, True])
def test_indexed(forget_length):
    original = ak.contents.IndexedArray(
        ak.index.Index64([0, 2]),
        ak.contents.ListOffsetArray(
            ak.index.Index64([0, 3, 6, 9]),
            ak.contents.NumpyArray(numpy.arange(9, dtype=numpy.int64)),
        ),
    )
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(
        array,
        ak.types.from_datashape(
            'var * int64[parameters={"key": "value"}]', highlevel=False
        ),
    )
    expected = ak.contents.IndexedArray(
        ak.index.Index64([0, 2]),
        ak.contents.ListOffsetArray(
            ak.index.Index(numpy.array([0, 3, 6, 9], dtype=numpy.int64)),
            ak.contents.NumpyArray(
                numpy.array([0, 1, 2, 3, 4, 5, 6, 7, 8], dtype=numpy.int64),
                parameters={"key": "value"},
            ),
        ),
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    original = ak.contents.IndexedArray(
        ak.index.Index64([0, 2]),
        ak.contents.ListOffsetArray(
            ak.index.Index64([0, 3, 6, 9]),
            ak.contents.NumpyArray(numpy.arange(9, dtype=numpy.int64)),
        ),
    )
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(
        array, ak.types.from_datashape("var * float32", highlevel=False)
    )
    expected = ak.contents.ListOffsetArray(
        ak.index.Index64([0, 3, 6]),
        ak.contents.NumpyArray(numpy.array([0, 1, 2, 6, 7, 8], dtype=numpy.float32)),
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)


@pytest.mark.parametrize("forget_length", [False, True])
def test_unknown(forget_length):
    original = ak.contents.EmptyArray()
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(array, "unknown")
    expected = ak.contents.EmptyArray().to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    original = ak.contents.EmptyArray()
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(array, "int64")
    expected = ak.contents.NumpyArray(numpy.empty(0, numpy.int64)).to_typetracer(
        forget_length
    )
    assert result.layout.is_equal_to(expected)

    original = ak.contents.NumpyArray(numpy.empty(0, numpy.int64))
    array = original.to_typetracer(forget_length)
    with pytest.raises(
        TypeError, match=r"cannot convert non-EmptyArray layouts to a bare UnknownType"
    ):
        ak.enforce_type(array, "unknown")

    original = ak.contents.NumpyArray(numpy.empty(0, numpy.int64))
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(array, "?unknown")
    expected = ak.contents.IndexedOptionArray(
        ak.index.Index64([]), ak.contents.EmptyArray()
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)


@pytest.mark.parametrize("forget_length", [False, True])
def test_misc(forget_length):
    ## option → option (inside indexed)
    original = ak.to_layout([{"x": [1, 2, None]}, None, {"x": [3, 4, None]}])[
        [0, 2], :2
    ]
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(
        array,
        """
    ?{
        x: var * ?float32
    }
        """,
    )
    expected = ak.contents.IndexedOptionArray(
        ak.index.Index64([0, 1]),
        ak.contents.RecordArray(
            [
                ak.contents.ListOffsetArray(
                    ak.index.Index64([0, 2, 4]),
                    ak.contents.IndexedOptionArray(
                        ak.index.Index64([0, 1, 2, 3]),
                        ak.contents.NumpyArray(
                            numpy.array([1, 2, 3, 4], dtype=numpy.float32)
                        ),
                    ),
                )
            ],
            ["x"],
        ),
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    ## no option → option (inside indexed)
    original = ak.to_layout(
        [
            {"x": [1, 2, None]},
            {"x": [9, 9, None]},
            {"x": [3, 4, None]},
            {"x": [8, 8, None]},
        ]
    )[[0, 2], :2]
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(
        array,
        """
    ?{
        x: var * ?float32
    }
        """,
    )
    expected = ak.contents.UnmaskedArray(
        ak.contents.RecordArray(
            [
                ak.contents.ListOffsetArray(
                    ak.index.Index64([0, 2, 4]),
                    ak.contents.IndexedOptionArray(
                        ak.index.Index64([0, 1, 2, 3]),
                        ak.contents.NumpyArray(
                            numpy.array([1, 2, 3, 4], dtype=numpy.float32)
                        ),
                    ),
                )
            ],
            ["x"],
        ),
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)

    ## option (indexed) list of union → option list of no union (project)
    original = ak.to_layout([[1, "hi", "bye"], None])[[0, 1], 1:2]
    array = original.to_typetracer(forget_length)
    result = ak.enforce_type(
        array, ak.types.from_datashape("?var * string", highlevel=False)
    )
    expected = ak.contents.IndexedOptionArray(
        ak.index.Index64([0, -1]),
        ak.contents.ListOffsetArray(
            ak.index.Index64([0, 1]),
            ak.contents.IndexedArray(
                ak.index.Index64([0]),
                ak.contents.ListOffsetArray(
                    offsets=ak.index.Index64([0, 2, 5]),
                    content=ak.contents.NumpyArray(
                        numpy.array([104, 105, 98, 121, 101], dtype=numpy.uint8),
                        parameters={"__array__": "char"},
                    ),
                    parameters={"__array__": "string"},
                ),
            ),
        ),
    ).to_typetracer(forget_length)
    assert result.layout.is_equal_to(expected)



from __future__ import annotations

import numpy as np

import awkward as ak


def test_ak_full_like_with_broadcasting():
    a = ak.Array(np.ones((2, 2)))
    b = ak.full_like(a, fill_value=np.array([2.0, 3.0]))

    assert ak.to_list(b) == [[2.0, 3.0], [2.0, 3.0]]



from __future__ import annotations

import awkward as ak

fromdict = {
    "class": "RecordArray",
    "fields": ["muon", "jet"],
    "contents": [
        {
            "class": "ListOffsetArray",
            "offsets": "i64",
            "content": {
                "class": "RecordArray",
                "fields": ["pt", "eta", "phi", "crossref"],
                "contents": [
                    {
                        "class": "NumpyArray",
                        "primitive": "int64",
                        "inner_shape": [],
                        "parameters": {},
                        "form_key": "muon_pt!",
                    },
                    {
                        "class": "NumpyArray",
                        "primitive": "int64",
                        "inner_shape": [],
                        "parameters": {},
                        "form_key": "muon_eta!",
                    },
                    {
                        "class": "NumpyArray",
                        "primitive": "int64",
                        "inner_shape": [],
                        "parameters": {},
                        "form_key": "muon_phi!",
                    },
                    {
                        "class": "ListOffsetArray",
                        "offsets": "i64",
                        "content": {
                            "class": "NumpyArray",
                            "primitive": "int64",
                            "inner_shape": [],
                            "parameters": {},
                            "form_key": "muon_crossref_content!",
                        },
                        "parameters": {},
                        "form_key": "muon_crossref_index!",
                    },
                ],
                "parameters": {},
                "form_key": "muon_record!",
            },
            "parameters": {},
            "form_key": "muon_list!",
        },
        {
            "class": "ListOffsetArray",
            "offsets": "i64",
            "content": {
                "class": "RecordArray",
                "fields": [
                    "pt",
                    "eta",
                    "phi",
                    "crossref",
                    "thing1",
                ],
                "contents": [
                    {
                        "class": "NumpyArray",
                        "primitive": "int64",
                        "inner_shape": [],
                        "parameters": {},
                        "form_key": "jet_pt!",
                    },
                    {
                        "class": "NumpyArray",
                        "primitive": "int64",
                        "inner_shape": [],
                        "parameters": {},
                        "form_key": "jet_eta!",
                    },
                    {
                        "class": "NumpyArray",
                        "primitive": "int64",
                        "inner_shape": [],
                        "parameters": {},
                        "form_key": "jet_phi!",
                    },
                    {
                        "class": "ListOffsetArray",
                        "offsets": "i64",
                        "content": {
                            "class": "NumpyArray",
                            "primitive": "int64",
                            "inner_shape": [],
                            "parameters": {},
                            "form_key": "jet_crossref_content!",
                        },
                        "parameters": {},
                        "form_key": "jet_crossref_index!",
                    },
                    {
                        "class": "NumpyArray",
                        "primitive": "int64",
                        "inner_shape": [],
                        "parameters": {},
                        "form_key": "jet_thing1!",
                    },
                ],
                "parameters": {},
                "form_key": "jet_record!",
            },
            "parameters": {},
            "form_key": "jet_list!",
        },
    ],
    "parameters": {},
    "form_key": "outer!",
}

form = ak.forms.from_dict(fromdict)
ttlayout, report = ak.typetracer.typetracer_with_report(form)
ttarray = ak.Array(ttlayout)


def test_where():
    ak.where(abs(ttarray.jet.eta) < 1.0, 0.000511, ttarray.jet.thing1)


def test_maybe_where():
    maybe = ak.firsts(ttarray)
    ak.where(abs(maybe.jet.eta) < 1.0, 0.000511, maybe.jet.thing1)


def test_varmaybe_where():
    varmaybe = ak.pad_none(ttarray, 3)
    ak.where(abs(varmaybe.jet.eta) < 1.0, 0.000511, varmaybe.jet.thing1)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import pytest

import awkward as ak
import awkward._connect.cling
import awkward._lookup

ROOT = pytest.importorskip("ROOT")

compiler = ROOT.gInterpreter.Declare


# ---------------------------------------------------------------------
def add_column(rdf: ROOT.RDataFrame, arr_val: np.ndarray, name: str):
    v_col_org = rdf.GetColumnNames()
    l_col_org = [name.c_str() for name in v_col_org]
    l_col = []

    for col in l_col_org:
        l_col.append(col)

    data = ak.from_rdataframe(rdf, columns=l_col)
    d_data = {col: data[col] for col in l_col}

    d_data[name] = arr_val

    rdf = ak.to_rdataframe(d_data)

    return rdf


# ---------------------------------------------------------------------
def test_add_column():
    ak_array_x = ak.Array([1, 2, 3])
    ak_array_y = ak.Array([4, 5, 6])

    rdf = ak.to_rdataframe({"x": ak_array_x, "y": ak_array_y})

    rdf = rdf.Define("z", "ROOT::RVec<int>({1, 2, 3})")
    rdf = rdf.Define("w", "true")

    arr_val = ak.Array([10, 20, 30])

    rdf = add_column(rdf, arr_val, "values")

    rdf.Display().Print()


# ---------------------------------------------------------------------
def add_numpy_column(rdf: ROOT.RDataFrame, arr_val: np.ndarray, name: str):
    v_col_org = rdf.GetColumnNames()
    l_col_org = [name.c_str() for name in v_col_org]
    l_col = []

    for col in l_col_org:
        l_col.append(col)

    data = ak.from_rdataframe(rdf, columns=l_col)
    d_data = {col: data[col] for col in l_col}

    d_data[name] = arr_val

    rdf = ak.to_rdataframe(d_data)

    return rdf


# ---------------------------------------------------------------------
def test_add_numpy():
    d_data = {
        "x": np.array([1, 2, 3]),
        "y": np.array([4, 5, 6]),
    }

    rdf = ROOT.RDF.FromNumpy(d_data)
    rdf = rdf.Define("z", "ROOT::RVec<int>({1, 2, 3})")
    rdf = rdf.Define("w", "true")

    arr_val = np.array([10, 20, 30])

    rdf = add_numpy_column(rdf, arr_val, "values")

    rdf.Display().Print()



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/master/LICENSE

from __future__ import annotations

import numpy as np

import awkward as ak

to_list = ak.operations.to_list


def test_bool_RecordArray():
    array = ak.highlevel.Array(
        [
            {"x": True, "y": [True]},
            {"x": False, "y": [True, False]},
        ]
    )

    assert ak._do.is_unique(array.layout) is False
    assert ak._do.is_unique(array["x"].layout) is True
    assert ak._do.is_unique(array["y"].layout) is False


def test_bool_UnionArray():
    content1 = ak.operations.from_iter([[], [False], [True, True]], highlevel=False)
    content2 = ak.operations.from_iter([[False, False], [True], []], highlevel=False)
    tags = ak.index.Index8(np.array([0, 1, 0, 1, 0, 1], dtype=np.int8))
    index = ak.index.Index64(np.array([0, 0, 1, 1, 2, 2], dtype=np.int64))
    array = ak.contents.UnionArray.simplified(tags, index, [content1, content2])

    assert to_list(array) == [
        [],
        [False, False],
        [False],
        [True],
        [True, True],
        [],
    ]
    assert ak._do.is_unique(array) is False
    assert to_list(ak._do.unique(array, axis=None)) == [False, True]
    assert to_list(ak._do.unique(array, axis=-1)) == [
        [],
        [False],
        [False],
        [True],
        [True],
        [],
    ]


def test_bool_IndexedArray():
    content = ak.from_iter([[True], [False]], highlevel=False)
    index = ak.index.Index64(np.array([1, 0], dtype=np.int64))
    indexedarray = ak.contents.IndexedArray(index, content)

    assert ak._do.is_unique(indexedarray) is True

    listoffsetarray = ak.operations.from_iter([[True], [False]], highlevel=False)

    index = ak.index.Index64(np.array([1, 0], dtype=np.int64))
    indexedarray = ak.contents.IndexedArray(index, listoffsetarray)
    assert to_list(indexedarray) == [
        [False],
        [True],
    ]
    assert ak._do.is_unique(indexedarray) is True

    assert ak._do.is_unique(indexedarray) is True
    assert to_list(ak._do.unique(indexedarray)) == [
        False,
        True,
    ]
    assert to_list(ak._do.unique(indexedarray, axis=-1)) == [
        [False],
        [True],
    ]


def test_bool_subranges_equal():
    array = ak.contents.NumpyArray(
        np.array(
            [
                [True, False, True, True, False],
                [False, False, True, True, True],
                [False, True, False, True, True],
            ]
        )
    )

    starts = ak.index.Index64(np.array([0, 5, 10]))
    stops = ak.index.Index64(np.array([5, 10, 15]))

    result = ak.sort(array, axis=-1, highlevel=False).content._subranges_equal(
        starts, stops, 15
    )
    assert result is True

    starts = ak.index.Index64(np.array([0, 7]))
    stops = ak.index.Index64(np.array([7, 15]))

    assert (
        ak.sort(array, axis=-1, highlevel=False).content._subranges_equal(
            starts, stops, 15
        )
        is False
    )

    starts = ak.index.Index64(np.array([0]))
    stops = ak.index.Index64(np.array([15]))

    assert (
        ak.sort(array, axis=-1, highlevel=False).content._subranges_equal(
            starts, stops, 15
        )
        is False
    )



# AUTO GENERATED ON 2026-02-08 AT 18:23:02
# DO NOT EDIT BY HAND!
#
# To regenerate file, run
#
#     python dev/generate-tests.py
#

# fmt: off

import cupy
import cupy.testing as cpt
import numpy as np
import pytest

import awkward as ak
import awkward._connect.cuda as ak_cu
from awkward._backends.cupy import CupyBackend

cupy_backend = CupyBackend.instance()

def test_cudaawkward_RegularArray_getitem_next_array_regularize_64_1():
    toarray = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromarray = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    lenarray = 3
    size = 3
    funcC = cupy_backend['awkward_RegularArray_getitem_next_array_regularize', cupy.int64, cupy.int64]
    funcC(toarray, fromarray, lenarray, size)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_toarray = [1, 1, 1]
    cpt.assert_allclose(toarray[:len(pytest_toarray)], cupy.array(pytest_toarray))

def test_cudaawkward_RegularArray_getitem_next_array_regularize_64_2():
    toarray = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromarray = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    lenarray = 3
    size = 2
    funcC = cupy_backend['awkward_RegularArray_getitem_next_array_regularize', cupy.int64, cupy.int64]
    funcC(toarray, fromarray, lenarray, size)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_toarray = [1, 1, 1]
    cpt.assert_allclose(toarray[:len(pytest_toarray)], cupy.array(pytest_toarray))

def test_cudaawkward_RegularArray_getitem_next_array_regularize_64_3():
    toarray = cupy.array([123], dtype=cupy.int64)
    fromarray = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    lenarray = 3
    size = 1
    funcC = cupy_backend['awkward_RegularArray_getitem_next_array_regularize', cupy.int64, cupy.int64]

def test_cudaawkward_RegularArray_getitem_next_array_regularize_64_4():
    toarray = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromarray = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    lenarray = 3
    size = 2
    funcC = cupy_backend['awkward_RegularArray_getitem_next_array_regularize', cupy.int64, cupy.int64]
    funcC(toarray, fromarray, lenarray, size)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_toarray = [1, 1, 1]
    cpt.assert_allclose(toarray[:len(pytest_toarray)], cupy.array(pytest_toarray))

def test_cudaawkward_RegularArray_getitem_next_array_regularize_64_5():
    toarray = cupy.array([123], dtype=cupy.int64)
    fromarray = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    lenarray = 3
    size = 0
    funcC = cupy_backend['awkward_RegularArray_getitem_next_array_regularize', cupy.int64, cupy.int64]

def test_cudaawkward_RegularArray_getitem_next_array_regularize_64_6():
    toarray = cupy.array([123, 123], dtype=cupy.int64)
    fromarray = cupy.array([2, 3, 3, 4, 5, 5, 5, 5, 5, 6, 7, 8, 10, 11], dtype=cupy.int64)
    lenarray = 3
    size = 3
    funcC = cupy_backend['awkward_RegularArray_getitem_next_array_regularize', cupy.int64, cupy.int64]

def test_cudaawkward_RegularArray_getitem_next_array_regularize_64_7():
    toarray = cupy.array([123], dtype=cupy.int64)
    fromarray = cupy.array([2, 3, 3, 4, 5, 5, 5, 5, 5, 6, 7, 8, 10, 11], dtype=cupy.int64)
    lenarray = 3
    size = 2
    funcC = cupy_backend['awkward_RegularArray_getitem_next_array_regularize', cupy.int64, cupy.int64]

def test_cudaawkward_RegularArray_getitem_next_array_regularize_64_8():
    toarray = cupy.array([123], dtype=cupy.int64)
    fromarray = cupy.array([2, 3, 3, 4, 5, 5, 5, 5, 5, 6, 7, 8, 10, 11], dtype=cupy.int64)
    lenarray = 3
    size = 1
    funcC = cupy_backend['awkward_RegularArray_getitem_next_array_regularize', cupy.int64, cupy.int64]

def test_cudaawkward_RegularArray_getitem_next_array_regularize_64_9():
    toarray = cupy.array([123], dtype=cupy.int64)
    fromarray = cupy.array([2, 3, 3, 4, 5, 5, 5, 5, 5, 6, 7, 8, 10, 11], dtype=cupy.int64)
    lenarray = 3
    size = 2
    funcC = cupy_backend['awkward_RegularArray_getitem_next_array_regularize', cupy.int64, cupy.int64]

def test_cudaawkward_RegularArray_getitem_next_array_regularize_64_10():
    toarray = cupy.array([123], dtype=cupy.int64)
    fromarray = cupy.array([2, 3, 3, 4, 5, 5, 5, 5, 5, 6, 7, 8, 10, 11], dtype=cupy.int64)
    lenarray = 3
    size = 0
    funcC = cupy_backend['awkward_RegularArray_getitem_next_array_regularize', cupy.int64, cupy.int64]

def test_cudaawkward_RegularArray_getitem_next_array_regularize_64_11():
    toarray = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromarray = cupy.array([2, 1, 0, 1, 2, 0, 1, 2, 2, 2, 1, 2, 1, 0, 0, 0, 0], dtype=cupy.int64)
    lenarray = 3
    size = 3
    funcC = cupy_backend['awkward_RegularArray_getitem_next_array_regularize', cupy.int64, cupy.int64]
    funcC(toarray, fromarray, lenarray, size)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_toarray = [2, 1, 0]
    cpt.assert_allclose(toarray[:len(pytest_toarray)], cupy.array(pytest_toarray))

def test_cudaawkward_RegularArray_getitem_next_array_regularize_64_12():
    toarray = cupy.array([123], dtype=cupy.int64)
    fromarray = cupy.array([2, 1, 0, 1, 2, 0, 1, 2, 2, 2, 1, 2, 1, 0, 0, 0, 0], dtype=cupy.int64)
    lenarray = 3
    size = 2
    funcC = cupy_backend['awkward_RegularArray_getitem_next_array_regularize', cupy.int64, cupy.int64]

def test_cudaawkward_RegularArray_getitem_next_array_regularize_64_13():
    toarray = cupy.array([123], dtype=cupy.int64)
    fromarray = cupy.array([2, 1, 0, 1, 2, 0, 1, 2, 2, 2, 1, 2, 1, 0, 0, 0, 0], dtype=cupy.int64)
    lenarray = 3
    size = 1
    funcC = cupy_backend['awkward_RegularArray_getitem_next_array_regularize', cupy.int64, cupy.int64]

def test_cudaawkward_RegularArray_getitem_next_array_regularize_64_14():
    toarray = cupy.array([123], dtype=cupy.int64)
    fromarray = cupy.array([2, 1, 0, 1, 2, 0, 1, 2, 2, 2, 1, 2, 1, 0, 0, 0, 0], dtype=cupy.int64)
    lenarray = 3
    size = 2
    funcC = cupy_backend['awkward_RegularArray_getitem_next_array_regularize', cupy.int64, cupy.int64]

def test_cudaawkward_RegularArray_getitem_next_array_regularize_64_15():
    toarray = cupy.array([123], dtype=cupy.int64)
    fromarray = cupy.array([2, 1, 0, 1, 2, 0, 1, 2, 2, 2, 1, 2, 1, 0, 0, 0, 0], dtype=cupy.int64)
    lenarray = 3
    size = 0
    funcC = cupy_backend['awkward_RegularArray_getitem_next_array_regularize', cupy.int64, cupy.int64]

def test_cudaawkward_RegularArray_getitem_next_array_regularize_64_16():
    toarray = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromarray = cupy.array([1, 0, 2, 3, 1, 2, 0, 0, 1, 1, 2, 3, 1, 2, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    lenarray = 3
    size = 3
    funcC = cupy_backend['awkward_RegularArray_getitem_next_array_regularize', cupy.int64, cupy.int64]
    funcC(toarray, fromarray, lenarray, size)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_toarray = [1, 0, 2]
    cpt.assert_allclose(toarray[:len(pytest_toarray)], cupy.array(pytest_toarray))

def test_cudaawkward_RegularArray_getitem_next_array_regularize_64_17():
    toarray = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromarray = cupy.array([1, 0, 2, 3, 1, 2, 0, 0, 1, 1, 2, 3, 1, 2, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    lenarray = 3
    size = 2
    funcC = cupy_backend['awkward_RegularArray_getitem_next_array_regularize', cupy.int64, cupy.int64]

def test_cudaawkward_RegularArray_getitem_next_array_regularize_64_18():
    toarray = cupy.array([123], dtype=cupy.int64)
    fromarray = cupy.array([1, 0, 2, 3, 1, 2, 0, 0, 1, 1, 2, 3, 1, 2, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    lenarray = 3
    size = 1
    funcC = cupy_backend['awkward_RegularArray_getitem_next_array_regularize', cupy.int64, cupy.int64]

def test_cudaawkward_RegularArray_getitem_next_array_regularize_64_19():
    toarray = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromarray = cupy.array([1, 0, 2, 3, 1, 2, 0, 0, 1, 1, 2, 3, 1, 2, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    lenarray = 3
    size = 2
    funcC = cupy_backend['awkward_RegularArray_getitem_next_array_regularize', cupy.int64, cupy.int64]

def test_cudaawkward_RegularArray_getitem_next_array_regularize_64_20():
    toarray = cupy.array([123], dtype=cupy.int64)
    fromarray = cupy.array([1, 0, 2, 3, 1, 2, 0, 0, 1, 1, 2, 3, 1, 2, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    lenarray = 3
    size = 0
    funcC = cupy_backend['awkward_RegularArray_getitem_next_array_regularize', cupy.int64, cupy.int64]

def test_cudaawkward_RegularArray_getitem_next_array_regularize_64_21():
    toarray = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromarray = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    lenarray = 3
    size = 3
    funcC = cupy_backend['awkward_RegularArray_getitem_next_array_regularize', cupy.int64, cupy.int64]
    funcC(toarray, fromarray, lenarray, size)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_toarray = [0, 0, 0]
    cpt.assert_allclose(toarray[:len(pytest_toarray)], cupy.array(pytest_toarray))

def test_cudaawkward_RegularArray_getitem_next_array_regularize_64_22():
    toarray = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromarray = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    lenarray = 3
    size = 2
    funcC = cupy_backend['awkward_RegularArray_getitem_next_array_regularize', cupy.int64, cupy.int64]
    funcC(toarray, fromarray, lenarray, size)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_toarray = [0, 0, 0]
    cpt.assert_allclose(toarray[:len(pytest_toarray)], cupy.array(pytest_toarray))

def test_cudaawkward_RegularArray_getitem_next_array_regularize_64_23():
    toarray = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromarray = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    lenarray = 3
    size = 1
    funcC = cupy_backend['awkward_RegularArray_getitem_next_array_regularize', cupy.int64, cupy.int64]
    funcC(toarray, fromarray, lenarray, size)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_toarray = [0, 0, 0]
    cpt.assert_allclose(toarray[:len(pytest_toarray)], cupy.array(pytest_toarray))

def test_cudaawkward_RegularArray_getitem_next_array_regularize_64_24():
    toarray = cupy.array([123, 123, 123], dtype=cupy.int64)
    fromarray = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    lenarray = 3
    size = 2
    funcC = cupy_backend['awkward_RegularArray_getitem_next_array_regularize', cupy.int64, cupy.int64]
    funcC(toarray, fromarray, lenarray, size)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_toarray = [0, 0, 0]
    cpt.assert_allclose(toarray[:len(pytest_toarray)], cupy.array(pytest_toarray))

def test_cudaawkward_RegularArray_getitem_next_array_regularize_64_25():
    toarray = cupy.array([123], dtype=cupy.int64)
    fromarray = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    lenarray = 3
    size = 0
    funcC = cupy_backend['awkward_RegularArray_getitem_next_array_regularize', cupy.int64, cupy.int64]




from __future__ import annotations

import numpy as np

import awkward as ak


def test():
    ak.concatenate([ak.Array([i])[:, np.newaxis] for i in range(127)], axis=1)

    ak.concatenate([ak.Array([i])[:, np.newaxis] for i in range(128)], axis=1)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward._connect.numpy import UNSUPPORTED
from awkward._dispatch import high_level_function
from awkward._layout import HighLevelContext
from awkward._namedaxis import (
    _get_named_axis,
    _keep_named_axis,
    _named_axis_to_positional_axis,
    _remove_named_axis,
)
from awkward._nplikes.numpy_like import NumpyMetadata
from awkward._regularize import regularize_axis

__all__ = ("argmin", "nanargmin")

np = NumpyMetadata.instance()


@high_level_function()
def argmin(
    array,
    axis=None,
    *,
    keepdims=False,
    mask_identity=True,
    highlevel=True,
    behavior=None,
    attrs=None,
):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        axis (None or int or str): If None, combine all values from the array into
            a single scalar result; if an int, group by that axis: `0` is the
            outermost, `1` is the first level of nested lists, etc., and
            negative `axis` counts from the innermost: `-1` is the innermost,
            `-2` is the next level up, etc; if a str, it is interpreted as the
            name of the axis which maps to an int if named axes are present.
            Named axes are attached to an array using #ak.with_named_axis and
            removed with #ak.without_named_axis; also see the
            [Named axes user guide](../../user-guide/how-to-array-properties-named-axis.html).
        keepdims (bool): If False, this reducer decreases the number of
            dimensions by 1; if True, the reduced values are wrapped in a new
            length-1 dimension so that the result of this operation may be
            broadcasted with the original array.
        mask_identity (bool): If True, reducing over empty lists results in
            None (an option type); otherwise, reducing over empty lists
            results in the operation's identity.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        attrs (None or dict): Custom attributes for the output array, if
            high-level.

    Returns the index position of the minimum value in each group of elements
    from `array` (many types supported, including all Awkward Arrays and
    Records). The identity of minimization would be infinity, but argmin
    must return the position of the minimum element, which has no value for
    empty lists. Therefore, the identity should be masked: the argmin of
    an empty list is None. If `mask_identity=False`, the result would be `-1`,
    which is distinct from all valid index positions, but care should be taken
    that it is not misinterpreted as "the last element of the list."

    This operation is the same as NumPy's
    [argmin](https://docs.scipy.org/doc/numpy/reference/generated/numpy.argmin.html)
    if all lists at a given dimension have the same length and no None values,
    but it generalizes to cases where they do not.

    See #ak.sum for a more complete description of nested list and missing
    value (None) handling in reducers.

    See also #ak.nanargmin.
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, axis, keepdims, mask_identity, highlevel, behavior, attrs)


@high_level_function()
def nanargmin(
    array,
    axis=None,
    *,
    keepdims=False,
    mask_identity=True,
    highlevel=True,
    behavior=None,
    attrs=None,
):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        axis (None or int or str): If None, combine all values from the array into
            a single scalar result; if an int, group by that axis: `0` is the
            outermost, `1` is the first level of nested lists, etc., and
            negative `axis` counts from the innermost: `-1` is the innermost,
            `-2` is the next level up, etc; if a str, it is interpreted as the
            name of the axis which maps to an int if named axes are present.
            Named axes are attached to an array using #ak.with_named_axis and
            removed with #ak.without_named_axis; also see the
            [Named axes user guide](../../user-guide/how-to-array-properties-named-axis.html).
        keepdims (bool): If False, this reducer decreases the number of
            dimensions by 1; if True, the reduced values are wrapped in a new
            length-1 dimension so that the result of this operation may be
            broadcasted with the original array.
        mask_identity (bool): If True, reducing over empty lists results in
            None (an option type); otherwise, reducing over empty lists
            results in the operation's identity.
        mask_identity (bool): If True, reducing over empty lists results in
            None (an option type); otherwise, reducing over empty lists
            results in the operation's identity.

    Like #ak.argmin, but treating NaN ("not a number") values as missing.

    Equivalent to

        ak.argmin(ak.nan_to_none(array))

    with all other arguments unchanged.

    See also #ak.argmin.
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(
        ak.operations.ak_nan_to_none._impl(array, True, behavior, None),
        axis,
        keepdims,
        mask_identity,
        highlevel,
        behavior,
        attrs,
    )


def _impl(array, axis, keepdims, mask_identity, highlevel, behavior, attrs):
    with HighLevelContext(behavior=behavior, attrs=attrs) as ctx:
        layout = ctx.unwrap(array, allow_record=False, primitive_policy="error")

    # Handle named axis
    named_axis = _get_named_axis(ctx)
    # Step 1: Normalize named axis to positional axis
    axis = _named_axis_to_positional_axis(named_axis, axis)
    # Step 2: propagate named axis from input to output,
    #   keepdims=True: use strategy "keep all" (see: awkward._namedaxis)
    #   keepdims=False: use strategy "remove one" (see: awkward._namedaxis)
    out_named_axis = _keep_named_axis(named_axis, None)
    if not keepdims:
        out_named_axis = _remove_named_axis(
            named_axis=out_named_axis,
            axis=axis,
            total=layout.minmax_depth[1],
        )

    axis = regularize_axis(axis, none_allowed=True)

    reducer = ak._reducers.ArgMin()

    out = ak._do.reduce(
        layout,
        reducer,
        axis=axis,
        mask=mask_identity,
        keepdims=keepdims,
        behavior=ctx.behavior,
    )

    wrapped_out = ctx.wrap(
        out,
        highlevel=highlevel,
        allow_other=True,
    )

    # propagate named axis to output
    return ak.operations.ak_with_named_axis._impl(
        wrapped_out,
        named_axis=out_named_axis,
        highlevel=highlevel,
        behavior=ctx.behavior,
        attrs=ctx.attrs,
    )


@ak._connect.numpy.implements("argmin")
def _nep_18_impl_argmin(a, axis=None, out=UNSUPPORTED, *, keepdims=False):
    return argmin(a, axis=axis, keepdims=keepdims)


@ak._connect.numpy.implements("nanargmin")
def _nep_18_impl_nanargmin(a, axis=None, out=UNSUPPORTED, *, keepdims=False):
    return nanargmin(a, axis=axis, keepdims=keepdims)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward._dispatch import high_level_function
from awkward._layout import HighLevelContext

__all__ = ("categories",)


@high_level_function()
def categories(array, highlevel=True, *, behavior=None, attrs=None):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        attrs (None or dict): Custom attributes for the output array, if
            high-level.

    If the `array` is categorical (contains #ak.contents.IndexedArray or
    #ak.contents.IndexedOptionArray labeled with parameter
    `"__array__" = "categorical"`), then this function returns its categories.

    See also #ak.is_categorical, #ak.str.to_categorical, #ak.from_categorical.
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, highlevel, behavior, attrs)


def _impl(array, highlevel, behavior, attrs):
    output = None

    def action(layout, **kwargs):
        nonlocal output

        if layout.parameter("__array__") == "categorical":
            output = layout.content
            return layout

        else:
            return None

    with HighLevelContext(behavior=behavior, attrs=attrs) as ctx:
        layout = ctx.unwrap(array, allow_record=False, primitive_policy="error")

    ak._do.recursively_apply(layout, action)

    wrapped_out = ctx.wrap(output, highlevel=highlevel, allow_other=True)

    # propagate named axis from input to output,
    #   use strategy "drop all" (see: awkward._namedaxis)
    return ak.operations.ak_without_named_axis._impl(
        wrapped_out,
        highlevel=highlevel,
        behavior=ctx.behavior,
        attrs=ctx.attrs,
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak


def virtualize(array):
    form, length, container = ak.to_buffers(array)
    new_container = {k: lambda v=v: v for k, v in container.items()}
    return ak.from_buffers(form, length, new_container)


def test_merge_union_of_records():
    a1 = ak.Array([{"a": 1, "b": 2}])
    a2 = ak.Array([{"b": 3.3, "c": 4.4}])
    c = ak.concatenate((a1, a2))
    c = virtualize(c)

    assert c.tolist() == [{"a": 1, "b": 2}, {"b": 3.3, "c": 4.4}]

    assert str(c.type) == "2 * union[{a: int64, b: int64}, {b: float64, c: float64}]"

    d = ak.merge_union_of_records(c)

    assert d.tolist() == [{"a": 1, "b": 2, "c": None}, {"a": None, "b": 3.3, "c": 4.4}]

    assert str(d.type) == "2 * {a: ?int64, b: float64, c: ?float64}"


def test_merge_union_of_records_2():
    a1 = ak.Array([{"a": 1, "b": 2}])
    a2 = ak.Array([{"b": 3.3, "c": 4.4}, {"b": None, "c": None}])
    c = ak.concatenate((a1, a2))
    c = virtualize(c)

    assert c.tolist() == [
        {"a": 1, "b": 2},
        {"b": 3.3, "c": 4.4},
        {"b": None, "c": None},
    ]

    assert str(c.type) == "3 * union[{a: int64, b: int64}, {b: ?float64, c: ?float64}]"

    d = ak.merge_union_of_records(c)

    assert d.tolist() == [
        {"a": 1, "b": 2, "c": None},
        {"a": None, "b": 3.3, "c": 4.4},
        {"a": None, "b": None, "c": None},
    ]

    assert str(d.type) == "3 * {a: ?int64, b: ?float64, c: ?float64}"


def test_merge_union_of_records_3():
    a1 = ak.Array([[[[{"a": 1, "b": 2}]]]])
    a2 = ak.Array([[[[{"b": 3.3, "c": 4.4}]]]])
    c = ak.concatenate((a1, a2), axis=-1)
    c = virtualize(c)

    assert c.tolist() == [[[[{"a": 1, "b": 2}, {"b": 3.3, "c": 4.4}]]]]

    assert (
        str(c.type)
        == "1 * var * var * var * union[{a: int64, b: int64}, {b: float64, c: float64}]"
    )

    d = ak.merge_union_of_records(c, axis=-1)

    assert d.tolist() == [
        [[[{"a": 1, "b": 2, "c": None}, {"a": None, "b": 3.3, "c": 4.4}]]]
    ]

    assert str(d.type) == "1 * var * var * var * {a: ?int64, b: float64, c: ?float64}"


def test_merge_option_of_records():
    a = ak.Array([None, {"a": 1, "b": 2}])
    a = virtualize(a)

    assert str(a.type) == "2 * ?{a: int64, b: int64}"

    b = ak.merge_option_of_records(a)

    assert b.tolist() == [{"a": None, "b": None}, {"a": 1, "b": 2}]

    assert str(b.type) == "2 * {a: ?int64, b: ?int64}"


def test_merge_option_of_records_2():
    a = ak.Array([None, {"a": 1, "b": 2}, {"a": None, "b": None}])
    a = virtualize(a)

    assert str(a.type) == "3 * ?{a: ?int64, b: ?int64}"

    b = ak.merge_option_of_records(a)

    assert b.tolist() == [
        {"a": None, "b": None},
        {"a": 1, "b": 2},
        {"a": None, "b": None},
    ]

    assert str(b.type) == "3 * {a: ?int64, b: ?int64}"


def test_merge_option_of_records_3():
    a = ak.Array([[[[None, {"a": 1, "b": 2}]]]])
    a = virtualize(a)

    assert str(a.type) == "1 * var * var * var * ?{a: int64, b: int64}"

    b = ak.merge_option_of_records(a, axis=-1)

    assert b.tolist() == [[[[{"a": None, "b": None}, {"a": 1, "b": 2}]]]]

    assert str(b.type) == "1 * var * var * var * {a: ?int64, b: ?int64}"


def test_indexed():
    x = ak.to_layout([{"a": 1, "b": 2}])
    y = ak.contents.IndexedArray(
        ak.index.Index64([1]), ak.to_layout([{"c": 13, "b": 15}, {"c": 3, "b": 5}])
    )

    z = ak.concatenate((x, y))
    z = virtualize(z)

    assert z.tolist() == [
        {"a": 1, "b": 2},
        {"c": 3, "b": 5},
    ]
    assert z.type == ak.types.ArrayType(
        ak.types.UnionType(
            [
                ak.types.RecordType(
                    [ak.types.NumpyType("int64"), ak.types.NumpyType("int64")],
                    ["a", "b"],
                ),
                ak.types.RecordType(
                    [ak.types.NumpyType("int64"), ak.types.NumpyType("int64")],
                    ["c", "b"],
                ),
            ]
        ),
        2,
    )

    w = ak.merge_union_of_records(z)
    assert w.type == ak.types.ArrayType(
        ak.types.RecordType(
            [
                ak.types.OptionType(ak.types.NumpyType("int64")),
                ak.types.NumpyType("int64"),
                ak.types.OptionType(ak.types.NumpyType("int64")),
            ],
            ["a", "b", "c"],
        ),
        2,
    )
    assert w.tolist() == [{"a": 1, "b": 2, "c": None}, {"a": None, "b": 5, "c": 3}]


def test_option():
    x = ak.to_layout([{"a": 1, "b": 2}])
    y = ak.to_layout([{"c": 3, "b": 5}, None])

    z = ak.concatenate((x, y))
    z = virtualize(z)

    assert z.tolist() == [{"a": 1, "b": 2}, {"c": 3, "b": 5}, None]
    assert z.type == ak.types.ArrayType(
        ak.types.UnionType(
            [
                ak.types.OptionType(
                    ak.types.RecordType(
                        [ak.types.NumpyType("int64"), ak.types.NumpyType("int64")],
                        ["a", "b"],
                    )
                ),
                ak.types.OptionType(
                    ak.types.RecordType(
                        [ak.types.NumpyType("int64"), ak.types.NumpyType("int64")],
                        ["c", "b"],
                    )
                ),
            ]
        ),
        3,
    )

    w = ak.merge_union_of_records(z)
    assert w.type == ak.types.ArrayType(
        ak.types.OptionType(
            ak.types.RecordType(
                [
                    ak.types.OptionType(ak.types.NumpyType("int64")),
                    ak.types.NumpyType("int64"),
                    ak.types.OptionType(ak.types.NumpyType("int64")),
                ],
                ["a", "b", "c"],
            )
        ),
        3,
    )
    assert w.tolist() == [
        {"a": 1, "b": 2, "c": None},
        {"a": None, "b": 5, "c": 3},
        None,
    ]


def test_option_unmasked():
    x = ak.to_layout([{"a": 1, "b": 2}])
    y = ak.contents.UnmaskedArray(ak.to_layout([{"c": 3, "b": 5}]))

    z = ak.concatenate((x, y))
    z = virtualize(z)

    assert z.tolist() == [{"a": 1, "b": 2}, {"c": 3, "b": 5}]
    assert z.type == ak.types.ArrayType(
        ak.types.UnionType(
            [
                ak.types.OptionType(
                    ak.types.RecordType(
                        [ak.types.NumpyType("int64"), ak.types.NumpyType("int64")],
                        ["a", "b"],
                    )
                ),
                ak.types.OptionType(
                    ak.types.RecordType(
                        [ak.types.NumpyType("int64"), ak.types.NumpyType("int64")],
                        ["c", "b"],
                    )
                ),
            ]
        ),
        2,
    )

    w = ak.merge_union_of_records(z)
    assert w.type == ak.types.ArrayType(
        ak.types.OptionType(
            ak.types.RecordType(
                [
                    ak.types.OptionType(ak.types.NumpyType("int64")),
                    ak.types.NumpyType("int64"),
                    ak.types.OptionType(ak.types.NumpyType("int64")),
                ],
                ["a", "b", "c"],
            )
        ),
        2,
    )
    assert w.tolist() == [{"a": 1, "b": 2, "c": None}, {"a": None, "b": 5, "c": 3}]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import pytest

import awkward as ak


def test():
    arr = ak.Array([1])

    with pytest.raises(
        TypeError,
        match="'attrs' keys must be strings, got: 1",
    ):
        arr.attrs[1] = "foo"



from __future__ import annotations

import cupy as cp
import pytest

import awkward as ak

to_list = ak.operations.to_list


@pytest.fixture(scope="function", autouse=True)
def cleanup_cuda():
    yield
    try:
        cp.cuda.Device().synchronize()  # wait for all kernels
    except cp.cuda.runtime.CUDARuntimeError as e:
        print("GPU error during sync:", e)
    cp._default_memory_pool.free_all_blocks()


def test_0835_argmin_argmax_axis_None():
    array = ak.highlevel.Array(
        [
            [
                [2022, 2023, 2025],
                [],
                [2027, 2011],
                [2013],
            ],
            [],
            [[2017, 2019], [2023]],
        ],
    )
    cuda_array = ak.to_backend(array, "cuda")

    assert ak.operations.argmin(cuda_array) == 4
    assert ak.operations.argmax(cuda_array) == 3


def test_1106_argminmax_axis_None_missing_values():
    array = ak.highlevel.Array([1, 2, 3, None, 4])

    cuda_array = ak.to_backend(array, "cuda")

    assert ak.operations.argmax(cuda_array) == 4


def test_0070_argmin_and_argmax_jagged():
    v2_array = ak.operations.from_iter(
        [[2.2, 1.1, 3.3], [], [4.4, 5.5], [5.5], [-4.4, -5.5, -6.6]], highlevel=False
    )

    cuda_v2_array = ak.to_backend(v2_array, "cuda", highlevel=False)

    assert to_list(ak.argmin(cuda_v2_array, axis=1, highlevel=False)) == [
        1,
        None,
        0,
        0,
        2,
    ]
    assert (
        ak.argmin(cuda_v2_array.to_typetracer(), axis=1, highlevel=False).form
        == ak.argmin(cuda_v2_array, axis=1, highlevel=False).form
    )

    index2 = ak.index.Index64(cp.array([4, 3, 2, 1, 0], dtype=cp.int64))
    cuda_v2_array2 = ak.contents.IndexedArray(index2, cuda_v2_array)

    assert to_list(ak.argmin(cuda_v2_array2, axis=1, highlevel=False)) == [
        2,
        0,
        0,
        None,
        1,
    ]
    assert (
        ak.argmin(cuda_v2_array2.to_typetracer(), axis=1, highlevel=False).form
        == ak.argmin(cuda_v2_array2, axis=1, highlevel=False).form
    )

    index3 = ak.index.Index64(cp.array([4, 3, -1, 4, 0], dtype=cp.int64))
    cuda_v2_array2 = ak.contents.IndexedOptionArray(index3, cuda_v2_array)

    assert to_list(ak.argmin(cuda_v2_array2, axis=1, highlevel=False)) == [
        2,
        0,
        None,
        2,
        1,
    ]
    assert (
        ak.argmin(cuda_v2_array2.to_typetracer(), axis=1, highlevel=False).form
        == ak.argmin(cuda_v2_array2, axis=1, highlevel=False).form
    )
    assert to_list(ak.argmin(cuda_v2_array2, axis=-1, highlevel=False)) == [
        2,
        0,
        None,
        2,
        1,
    ]
    assert (
        ak.argmin(cuda_v2_array2.to_typetracer(), axis=-1, highlevel=False).form
        == ak.argmin(cuda_v2_array2, axis=-1, highlevel=False).form
    )


def test_0070_argmin_and_argmax_missing():
    array = ak.operations.from_iter(
        [[[2.2, 1.1, 3.3]], [[]], [None, None, None], [[-4.4, -5.5, -6.6]]],
        highlevel=False,
    )

    cuda_array = ak.to_backend(array, "cuda", highlevel=False)

    assert to_list(ak.argmin(cuda_array, axis=2, highlevel=False)) == [
        [1],
        [None],
        [None, None, None],
        [2],
    ]
    assert (
        ak.argmin(cuda_array.to_typetracer(), axis=2, highlevel=False).form
        == ak.argmin(cuda_array, axis=2, highlevel=False).form
    )


def test_0115_generic_reducer_operation_ByteMaskedArray():
    content = ak.operations.from_iter(
        [
            [[1.1, 0.0, 2.2], [], [3.3, 4.4]],
            [],
            [[5.5]],
            [[6.6, 9.9, 8.8, 7.7]],
            [[], [12.2, 11.1, 10.0]],
        ],
        highlevel=False,
    )
    mask = ak.index.Index8(cp.array([0, 0, 1, 1, 0], dtype=cp.int8))
    content = ak.to_backend(content, "cuda", highlevel=False)

    cuda_v2_array = ak.contents.ByteMaskedArray(mask, content, valid_when=False)

    assert to_list(cuda_v2_array) == [
        [[1.1, 0.0, 2.2], [], [3.3, 4.4]],
        [],
        None,
        None,
        [[], [12.2, 11.1, 10.0]],
    ]
    assert to_list(ak.argmin(cuda_v2_array, axis=-1, highlevel=False)) == [
        [1, None, 0],
        [],
        None,
        None,
        [None, 2],
    ]
    assert (
        ak.argmin(cuda_v2_array.to_typetracer(), axis=-1, highlevel=False).form
        == ak.argmin(cuda_v2_array, axis=-1, highlevel=False).form
    )


@pytest.mark.parametrize(
    "func",
    [
        ak.argmin,
        ak.argmax,
    ],
)
def test_2754_highlevel_behavior_missing_reducers(func):
    behavior_1 = {"foo": "bar"}
    behavior_2 = {"baz": "bargh!"}

    array = ak.Array([[1, 2, 3, 4], [5], [10]])

    cuda_array = ak.to_backend(array, "cuda")

    assert isinstance(func(cuda_array, axis=1, highlevel=True), ak.Array)
    assert isinstance(func(cuda_array, axis=1, highlevel=False), ak.contents.Content)
    assert (
        func(
            ak.Array(cuda_array, behavior=behavior_1),
            axis=1,
            highlevel=True,
            behavior=behavior_2,
        ).behavior
        == behavior_2
    )
    assert (
        func(
            ak.Array(cuda_array, behavior=behavior_1),
            axis=1,
            highlevel=True,
        ).behavior
        == behavior_1
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak


def test_attrs():
    x = ak.Array([1], attrs={"me": "ta"})
    y = ak.Array([1], attrs={"da": "ta"})

    rec = ak.Array({"x": x, "y": y})
    assert rec.attrs == {"me": "ta", "da": "ta"}

    # overwrite attrs:
    rec = ak.Array({"x": x, "y": y}, attrs={"foo": "bar"})
    assert rec.attrs == {"foo": "bar"}


def test_behaviors():
    behavior = {"foo": object()}
    x = ak.Array([1], behavior=behavior)
    y = ak.Array([1], behavior=behavior)

    rec = ak.Array({"x": x, "y": y})
    assert rec.behavior == behavior

    # overwrite behavior:
    behavior2 = {"bar": object()}
    rec = ak.Array({"x": x, "y": y}, behavior=behavior2)
    assert rec.behavior == behavior2



from __future__ import annotations

import awkward as ak


def test():
    length0 = ak.Array([])
    assert repr(length0) == "<Array [] type='0 * unknown'>"
    assert length0._repr_mimebundle_() == {
        "text/html": "<pre>\nbackend: cpu\nnbytes: 0 B\ntype: 0 * unknown</pre>",
        "text/plain": "<Array [] type='0 * unknown'>",
    }



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numbers
import os
from collections.abc import Iterable, Sequence, Sized

from awkward._nplikes.numpy_like import NumpyMetadata
from awkward._typing import Any

np = NumpyMetadata.instance()


def is_file_path(x) -> bool:
    try:
        return os.path.isfile(x)
    except ValueError:
        return False


def is_sized_iterable(obj) -> bool:
    return isinstance(obj, Iterable) and isinstance(obj, Sized)


def is_integer(x) -> bool:
    return isinstance(x, numbers.Integral) and not isinstance(x, bool)


def is_array_like(x) -> bool:
    return hasattr(x, "shape") and hasattr(x, "dtype") and hasattr(x, "T")


def is_integer_like(x) -> bool:
    # Integral types
    if isinstance(x, numbers.Integral):
        return not isinstance(x, bool)
    # Scalar arrays
    elif is_array_like(x):
        return np.issubdtype(x.dtype, np.integer) and x.ndim == 0
    # Other things that support integers
    else:
        return hasattr(x, "__int__")


def is_non_string_like_iterable(obj) -> bool:
    return not isinstance(obj, (str, bytes)) and isinstance(obj, Iterable)


def is_non_string_like_sequence(obj) -> bool:
    return not isinstance(obj, (str, bytes)) and isinstance(obj, Sequence)


def regularize_axis(axis: Any, none_allowed: bool = True) -> int | None:
    """
    This function's main purpose is to convert [np,cp,...].array(0) to 0.
    """
    if is_integer_like(axis):
        regularized_axis = int(axis)
    else:
        regularized_axis = axis
    cond = is_integer(regularized_axis)
    msg = f"'axis' must be an integer, not {axis!r}"
    if none_allowed:
        cond = cond or regularized_axis is None
        msg = f"'axis' must be an integer or None, not {axis!r}"
    if not cond:
        raise TypeError(msg)
    return regularized_axis



from __future__ import annotations

import cupy as cp
import numpy as np
import pytest

import awkward as ak

to_list = ak.operations.to_list

content = ak.contents.NumpyArray(
    np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
)
starts1 = ak.index.IndexU32(np.array([0, 3, 3, 5, 6], np.uint32))
stops1 = ak.index.IndexU32(np.array([3, 3, 5, 6, 9], np.uint32))
offsets1 = ak.index.IndexU32(np.array([0, 3, 3, 5, 6, 9], np.uint32))
starts2 = ak.index.IndexU32(np.array([0, 2, 3, 3], np.uint32))
stops2 = ak.index.IndexU32(np.array([2, 3, 3, 5], np.uint32))
offsets2 = ak.index.IndexU32(np.array([0, 2, 3, 3, 5], np.uint32))


@pytest.fixture(scope="function", autouse=True)
def cleanup_cuda():
    yield
    try:
        cp.cuda.Device().synchronize()  # wait for all kernels
    except cp.cuda.runtime.CUDARuntimeError as e:
        print("GPU error during sync:", e)
    cp._default_memory_pool.free_all_blocks()


def tests_0020_support_unsigned_indexes_listarray_ellipsis():
    array1 = ak.contents.ListArray(starts1, stops1, content)
    array2 = ak.contents.ListArray(starts2, stops2, array1)

    cuda_array1 = ak.to_backend(array1, "cuda", highlevel=False)
    cuda_array2 = ak.to_backend(array2, "cuda", highlevel=False)

    assert to_list(cuda_array1[Ellipsis, 1:]) == [
        [2.2, 3.3],
        [],
        [5.5],
        [],
        [8.8, 9.9],
    ]
    assert (
        cuda_array1.to_typetracer()[Ellipsis, 1:].form == cuda_array1[Ellipsis, 1:].form
    )
    assert to_list(cuda_array2[Ellipsis, 1:]) == [
        [[2.2, 3.3], []],
        [[5.5]],
        [],
        [[], [8.8, 9.9]],
    ]
    assert (
        cuda_array2.to_typetracer()[Ellipsis, 1:].form == cuda_array2[Ellipsis, 1:].form
    )


def tests_0020_support_unsigned_indexes_listoffsetarray_ellipsis():
    array1 = ak.contents.ListOffsetArray(offsets1, content)
    array2 = ak.contents.ListOffsetArray(offsets2, array1)

    cuda_array1 = ak.to_backend(array1, "cuda", highlevel=False)
    cuda_array2 = ak.to_backend(array2, "cuda", highlevel=False)

    assert to_list(cuda_array1[Ellipsis, 1:]) == [
        [2.2, 3.3],
        [],
        [5.5],
        [],
        [8.8, 9.9],
    ]
    assert (
        cuda_array1.to_typetracer()[Ellipsis, 1:].form == cuda_array1[Ellipsis, 1:].form
    )
    assert to_list(cuda_array2[Ellipsis, 1:]) == [
        [[2.2, 3.3], []],
        [[5.5]],
        [],
        [[], [8.8, 9.9]],
    ]
    assert (
        cuda_array2.to_typetracer()[Ellipsis, 1:].form == cuda_array2[Ellipsis, 1:].form
    )


def tests_0020_support_unsigned_indexes_listarray_array_slice():
    array1 = ak.contents.ListArray(starts1, stops1, content)
    array2 = ak.contents.ListArray(starts2, stops2, array1)

    cuda_array2 = ak.to_backend(array2, "cuda", highlevel=False)

    assert to_list(cuda_array2[[0, 0, 1, 1, 1, 0]]) == [
        [[1.1, 2.2, 3.3], []],
        [[1.1, 2.2, 3.3], []],
        [[4.4, 5.5]],
        [[4.4, 5.5]],
        [[4.4, 5.5]],
        [[1.1, 2.2, 3.3], []],
    ]
    assert (
        cuda_array2.to_typetracer()[[0, 0, 1, 1, 1, 0]].form
        == cuda_array2[[0, 0, 1, 1, 1, 0]].form
    )

    assert to_list(cuda_array2[[0, 0, 1, 1, 1, 0], :]) == [
        [[1.1, 2.2, 3.3], []],
        [[1.1, 2.2, 3.3], []],
        [[4.4, 5.5]],
        [[4.4, 5.5]],
        [[4.4, 5.5]],
        [[1.1, 2.2, 3.3], []],
    ]
    assert (
        cuda_array2.to_typetracer()[[0, 0, 1, 1, 1, 0], :].form
        == cuda_array2[[0, 0, 1, 1, 1, 0], :].form
    )
    assert to_list(cuda_array2[[0, 0, 1, 1, 1, 0], :, 1:]) == [
        [[2.2, 3.3], []],
        [[2.2, 3.3], []],
        [[5.5]],
        [[5.5]],
        [[5.5]],
        [[2.2, 3.3], []],
    ]
    assert (
        cuda_array2.to_typetracer()[[0, 0, 1, 1, 1, 0], :, 1:].form
        == cuda_array2[[0, 0, 1, 1, 1, 0], :, 1:].form
    )


def tests_0020_support_unsigned_indexes_listoffsetarray_array_slice():
    array1 = ak.contents.ListOffsetArray(offsets1, content)
    array2 = ak.contents.ListOffsetArray(offsets2, array1)

    cuda_array2 = ak.to_backend(array2, "cuda", highlevel=False)

    assert to_list(cuda_array2[[0, 0, 1, 1, 1, 0]]) == [
        [[1.1, 2.2, 3.3], []],
        [[1.1, 2.2, 3.3], []],
        [[4.4, 5.5]],
        [[4.4, 5.5]],
        [[4.4, 5.5]],
        [[1.1, 2.2, 3.3], []],
    ]
    assert (
        cuda_array2.to_typetracer()[[0, 0, 1, 1, 1, 0]].form
        == cuda_array2[[0, 0, 1, 1, 1, 0]].form
    )
    assert to_list(cuda_array2[[0, 0, 1, 1, 1, 0], :]) == [
        [[1.1, 2.2, 3.3], []],
        [[1.1, 2.2, 3.3], []],
        [[4.4, 5.5]],
        [[4.4, 5.5]],
        [[4.4, 5.5]],
        [[1.1, 2.2, 3.3], []],
    ]
    assert (
        cuda_array2.to_typetracer()[[0, 0, 1, 1, 1, 0], :].form
        == cuda_array2[[0, 0, 1, 1, 1, 0], :].form
    )
    assert to_list(cuda_array2[[0, 0, 1, 1, 1, 0], :, 1:]) == [
        [[2.2, 3.3], []],
        [[2.2, 3.3], []],
        [[5.5]],
        [[5.5]],
        [[5.5]],
        [[2.2, 3.3], []],
    ]
    assert (
        cuda_array2.to_typetracer()[[0, 0, 1, 1, 1, 0], :, 1:].form
        == cuda_array2[[0, 0, 1, 1, 1, 0], :, 1:].form
    )


def tests_0020_support_unsigned_indexes_listarray_array():
    array1 = ak.contents.ListArray(starts1, stops1, content)

    cuda_array1 = ak.to_backend(array1, "cuda", highlevel=False)

    assert to_list(cuda_array1[cp.array([2, 0, 0, 1, -1])]) == [
        [4.4, 5.5],
        [1.1, 2.2, 3.3],
        [1.1, 2.2, 3.3],
        [],
        [7.7, 8.8, 9.9],
    ]
    assert (
        cuda_array1.to_typetracer()[cp.array([2, 0, 0, 1, -1])].form
        == cuda_array1[cp.array([2, 0, 0, 1, -1])].form
    )
    assert to_list(cuda_array1[cp.array([2, 0, 0, -1]), cp.array([1, 1, 0, 0])]) == [
        5.5,
        2.2,
        1.1,
        7.7,
    ]
    assert (
        cuda_array1.to_typetracer()[
            cp.array([2, 0, 0, -1]), cp.array([1, 1, 0, 0])
        ].form
        == cuda_array1[cp.array([2, 0, 0, -1]), cp.array([1, 1, 0, 0])].form
    )


def tests_0020_support_unsigned_indexes_listoffsetarray_array():
    array1 = ak.contents.ListOffsetArray(offsets1, content)

    cuda_array1 = ak.to_backend(array1, "cuda", highlevel=False)

    assert to_list(cuda_array1[cp.array([2, 0, 0, 1, -1])]) == [
        [4.4, 5.5],
        [1.1, 2.2, 3.3],
        [1.1, 2.2, 3.3],
        [],
        [7.7, 8.8, 9.9],
    ]
    assert (
        cuda_array1.to_typetracer()[cp.array([2, 0, 0, 1, -1])].form
        == cuda_array1[cp.array([2, 0, 0, 1, -1])].form
    )
    assert to_list(cuda_array1[cp.array([2, 0, 0, -1]), cp.array([1, 1, 0, 0])]) == [
        5.5,
        2.2,
        1.1,
        7.7,
    ]
    assert (
        cuda_array1.to_typetracer()[
            cp.array([2, 0, 0, -1]), cp.array([1, 1, 0, 0])
        ].form
        == cuda_array1[cp.array([2, 0, 0, -1]), cp.array([1, 1, 0, 0])].form
    )


def tests_0020_support_unsigned_indexes_listarray_slice():
    array1 = ak.contents.ListArray(starts1, stops1, content)
    array2 = ak.contents.ListArray(starts2, stops2, array1)

    cuda_array1 = ak.to_backend(array1, "cuda", highlevel=False)
    cuda_array2 = ak.to_backend(array2, "cuda", highlevel=False)

    assert to_list(cuda_array1[1:-1]) == [[], [4.4, 5.5], [6.6]]
    assert cuda_array1.to_typetracer()[1:-1].form == cuda_array1[1:-1].form
    assert to_list(cuda_array1[1:-1,]) == [[], [4.4, 5.5], [6.6]]
    assert cuda_array1.to_typetracer()[1:-1,].form == cuda_array1[1:-1,].form
    assert to_list(cuda_array2[1:-1]) == [[[4.4, 5.5]], []]
    assert cuda_array2.to_typetracer()[1:-1].form == cuda_array2[1:-1].form
    assert to_list(cuda_array2[1:-1,]) == [[[4.4, 5.5]], []]
    assert cuda_array2.to_typetracer()[1:-1,].form == cuda_array2[1:-1,].form


def tests_0020_support_unsigned_indexes_listoffsetarray_slice():
    array1 = ak.contents.ListOffsetArray(offsets1, content)
    array2 = ak.contents.ListOffsetArray(offsets2, array1)

    cuda_array1 = ak.to_backend(array1, "cuda", highlevel=False)
    cuda_array2 = ak.to_backend(array2, "cuda", highlevel=False)

    assert to_list(cuda_array1[1:-1]) == [[], [4.4, 5.5], [6.6]]
    assert cuda_array1.to_typetracer()[1:-1].form == cuda_array1[1:-1].form
    assert to_list(cuda_array1[1:-1,]) == [[], [4.4, 5.5], [6.6]]
    assert cuda_array1.to_typetracer()[1:-1,].form == cuda_array1[1:-1,].form
    assert to_list(cuda_array2[1:-1]) == [[[4.4, 5.5]], []]
    assert cuda_array2.to_typetracer()[1:-1].form == cuda_array2[1:-1].form
    assert to_list(cuda_array2[1:-1,]) == [[[4.4, 5.5]], []]
    assert cuda_array2.to_typetracer()[1:-1,].form == cuda_array2[1:-1,].form


def tests_0020_support_unsigned_indexes_listarray_slice_slice():
    array1 = ak.contents.ListArray(starts1, stops1, content)

    cuda_array1 = ak.to_backend(array1, "cuda", highlevel=False)

    assert to_list(cuda_array1[2:]) == [[4.4, 5.5], [6.6], [7.7, 8.8, 9.9]]
    assert cuda_array1.to_typetracer()[2:].form == cuda_array1[2:].form
    assert to_list(cuda_array1[2:, 1:]) == [[5.5], [], [8.8, 9.9]]
    assert cuda_array1.to_typetracer()[2:, 1:].form == cuda_array1[2:, 1:].form
    assert to_list(cuda_array1[2:, :-1]) == [[4.4], [], [7.7, 8.8]]
    assert cuda_array1.to_typetracer()[2:, :-1].form == cuda_array1[2:, :-1].form


def tests_0020_support_unsigned_indexes_listoffsetarray_slice_slice():
    array1 = ak.contents.ListOffsetArray(offsets1, content)

    cuda_array1 = ak.to_backend(array1, "cuda", highlevel=False)

    assert to_list(cuda_array1[2:]) == [[4.4, 5.5], [6.6], [7.7, 8.8, 9.9]]
    assert cuda_array1.to_typetracer()[2:].form == cuda_array1[2:].form
    assert to_list(cuda_array1[2:, 1:]) == [[5.5], [], [8.8, 9.9]]
    assert cuda_array1.to_typetracer()[2:, 1:].form == cuda_array1[2:, 1:].form
    assert to_list(cuda_array1[2:, :-1]) == [[4.4], [], [7.7, 8.8]]
    assert cuda_array1.to_typetracer()[2:, :-1].form == cuda_array1[2:, :-1].form


class ReversibleArray(ak.Array):
    def reversed(self):
        return self[..., ::-1]


def test_2549_list_nominal_type_class():
    behavior = {"reversible": ReversibleArray}
    reversible_array = ak.with_parameter(
        [[1, 2, 3], [4, 5, 6, 7], [8, 9]], "__list__", "reversible", behavior=behavior
    )

    cuda_reversible_array = ak.to_backend(reversible_array, "cuda")

    assert isinstance(cuda_reversible_array, ReversibleArray)
    assert cuda_reversible_array.to_list() == [[1, 2, 3], [4, 5, 6, 7], [8, 9]]
    assert cuda_reversible_array.reversed().to_list() == [
        [3, 2, 1],
        [7, 6, 5, 4],
        [9, 8],
    ]


def test_2549_list_nominal_type_deep_class():
    behavior = {"reversible": ReversibleArray, ("*", "reversible"): ReversibleArray}
    reversible_array = ak.with_parameter(
        [[1, 2, 3], [4, 5, 6, 7], [8, 9]], "__list__", "reversible", behavior=behavior
    )
    outer_array = ak.Array(
        ak.contents.ListOffsetArray(
            ak.index.Index64([0, 2, 3, 3]), reversible_array.layout
        ),
        behavior=behavior,
    )

    cuda_outer_array = ak.to_backend(outer_array, "cuda")

    assert isinstance(cuda_outer_array, ReversibleArray)
    assert cuda_outer_array.to_list() == [[[1, 2, 3], [4, 5, 6, 7]], [[8, 9]], []]
    assert cuda_outer_array.reversed().to_list() == [
        [[3, 2, 1], [7, 6, 5, 4]],
        [[9, 8]],
        [],
    ]


def test_2549_list_nominal_type_ufunc():
    behavior = {"reversible": ReversibleArray}
    reversible_array = ak.with_parameter(
        [[1, 2, 3], [4, 5, 6, 7], [8, 9]], "__list__", "reversible", behavior=behavior
    )
    cuda_reversible_array = ak.to_backend(reversible_array, "cuda")

    assert isinstance(cuda_reversible_array, ReversibleArray)

    def reversible_add(x, y):
        return ak.with_parameter(x.reversed(), "__list__", None) + ak.with_parameter(
            y.reversed(), "__list__", None
        )

    ak.behavior[np.add, "reversible", "reversible"] = reversible_add

    assert (cuda_reversible_array + cuda_reversible_array).to_list() == [
        [6, 4, 2],
        [14, 12, 10, 8],
        [18, 16],
    ]
    with pytest.raises(TypeError, match=r"overloads for custom types"):
        cuda_reversible_array + ak.with_parameter(
            cuda_reversible_array, "__list__", "non-reversible"
        )


def test_0150_flatten_ListOffsetArray():
    array = ak.highlevel.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5]]).layout

    cuda_array = ak.to_backend(array, "cuda")

    assert ak.operations.to_list(ak.operations.flatten(cuda_array)) == [
        1.1,
        2.2,
        3.3,
        4.4,
        5.5,
    ]
    assert ak.operations.to_list(ak.operations.flatten(cuda_array[1:])) == [4.4, 5.5]

    array = ak.highlevel.Array(
        [[[0.0, 1.1, 2.2], [], [3.3, 4.4]], [], [[5.5]], [[], [6.6, 7.7, 8.8, 9.9]]]
    ).layout

    cuda_array = ak.to_backend(array, "cuda")

    assert ak.operations.to_list(ak.operations.flatten(cuda_array)) == [
        [0.0, 1.1, 2.2],
        [],
        [3.3, 4.4],
        [5.5],
        [],
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert ak.operations.to_list(ak.operations.flatten(cuda_array[1:])) == [
        [5.5],
        [],
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert ak.operations.to_list(ak.operations.flatten(cuda_array[:, 1:])) == [
        [],
        [3.3, 4.4],
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert ak.operations.to_list(ak.operations.flatten(cuda_array, axis=2)) == [
        [0.0, 1.1, 2.2, 3.3, 4.4],
        [],
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert ak.operations.to_list(ak.operations.flatten(cuda_array[1:], axis=2)) == [
        [],
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert ak.operations.to_list(ak.operations.flatten(cuda_array[:, 1:], axis=2)) == [
        [3.3, 4.4],
        [],
        [],
        [6.6, 7.7, 8.8, 9.9],
    ]

    array = ak.highlevel.Array(
        cp.arange(2 * 3 * 5 * 7).reshape(2, 3, 5, 7).tolist()
    ).layout

    cuda_array = ak.to_backend(array, "cuda")

    assert (
        ak.operations.to_list(ak.operations.flatten(cuda_array, axis=1))
        == cp.arange(2 * 3 * 5 * 7).reshape(2 * 3, 5, 7).tolist()
    )
    assert (
        ak.operations.to_list(ak.operations.flatten(cuda_array, axis=2))
        == cp.arange(2 * 3 * 5 * 7).reshape(2, 3 * 5, 7).tolist()
    )
    assert (
        ak.operations.to_list(ak.operations.flatten(cuda_array, axis=3))
        == cp.arange(2 * 3 * 5 * 7).reshape(2, 3, 5 * 7).tolist()
    )

    array = ak.highlevel.Array(
        ak.operations.from_iter(
            cp.arange(2 * 3 * 5 * 7).reshape(2, 3, 5, 7).tolist(), highlevel=False
        )
    ).layout

    cuda_array = ak.to_backend(array, "cuda")

    assert (
        ak.operations.to_list(ak.operations.flatten(cuda_array, axis=1))
        == cp.arange(2 * 3 * 5 * 7).reshape(2 * 3, 5, 7).tolist()
    )
    assert (
        ak.operations.to_list(ak.operations.flatten(cuda_array, axis=2))
        == cp.arange(2 * 3 * 5 * 7).reshape(2, 3 * 5, 7).tolist()
    )
    assert (
        ak.operations.to_list(ak.operations.flatten(cuda_array, axis=3))
        == cp.arange(2 * 3 * 5 * 7).reshape(2, 3, 5 * 7).tolist()
    )

    array = ak.highlevel.Array(cp.arange(2 * 3 * 5 * 7).reshape(2, 3, 5, 7)).layout

    cuda_array = ak.to_backend(array, "cuda")

    assert (
        ak.operations.to_list(ak.operations.flatten(cuda_array, axis=1))
        == cp.arange(2 * 3 * 5 * 7).reshape(2 * 3, 5, 7).tolist()
    )
    assert (
        ak.operations.to_list(ak.operations.flatten(cuda_array, axis=2))
        == cp.arange(2 * 3 * 5 * 7).reshape(2, 3 * 5, 7).tolist()
    )
    assert (
        ak.operations.to_list(ak.operations.flatten(cuda_array, axis=3))
        == cp.arange(2 * 3 * 5 * 7).reshape(2, 3, 5 * 7).tolist()
    )


def test_0150_flatten_IndexedArray():
    array = ak.highlevel.Array(
        [[1.1, 2.2, None, 3.3], None, [], None, [4.4, 5.5], None]
    ).layout

    cuda_array = ak.to_backend(array, "cuda")

    assert ak.operations.to_list(ak.operations.flatten(cuda_array)) == [
        1.1,
        2.2,
        None,
        3.3,
        4.4,
        5.5,
    ]
    assert ak.operations.to_list(ak.operations.flatten(cuda_array[1:])) == [4.4, 5.5]

    array = ak.highlevel.Array(
        [
            [[0.0, 1.1, 2.2], None, None, [3.3, 4.4]],
            [],
            [[5.5]],
            [[], [6.6, 7.7, 8.8, 9.9]],
        ]
    ).layout

    cuda_array = ak.to_backend(array, "cuda")

    assert ak.operations.to_list(ak.operations.flatten(cuda_array, axis=2)) == [
        [0.0, 1.1, 2.2, 3.3, 4.4],
        [],
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert ak.operations.to_list(ak.operations.flatten(cuda_array[1:], axis=2)) == [
        [],
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert ak.operations.to_list(ak.operations.flatten(cuda_array[:, 1:], axis=2)) == [
        [3.3, 4.4],
        [],
        [],
        [6.6, 7.7, 8.8, 9.9],
    ]

    array = ak.highlevel.Array(
        [
            [[0.0, 1.1, 2.2], [3.3, 4.4]],
            [],
            [[5.5]],
            None,
            None,
            [[], [6.6, 7.7, 8.8, 9.9]],
        ]
    ).layout

    cuda_array = ak.to_backend(array, "cuda")

    assert ak.operations.to_list(ak.operations.flatten(cuda_array, axis=2)) == [
        [0.0, 1.1, 2.2, 3.3, 4.4],
        [],
        [5.5],
        None,
        None,
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert ak.operations.to_list(ak.operations.flatten(cuda_array[1:], axis=2)) == [
        [],
        [5.5],
        None,
        None,
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert ak.operations.to_list(ak.operations.flatten(cuda_array[:, 1:], axis=2)) == [
        [3.3, 4.4],
        [],
        [],
        None,
        None,
        [6.6, 7.7, 8.8, 9.9],
    ]

    array = ak.highlevel.Array(
        [
            [[0.0, 1.1, None, 2.2], None, [], None, [3.3, 4.4]],
            None,
            [],
            [[5.5]],
            None,
            [[], [6.6, None, 7.7, 8.8, 9.9], None],
        ]
    ).layout

    cuda_array = ak.to_backend(array, "cuda")

    assert ak.operations.to_list(ak.operations.flatten(cuda_array)) == [
        [0.0, 1.1, None, 2.2],
        None,
        [],
        None,
        [3.3, 4.4],
        [5.5],
        [],
        [6.6, None, 7.7, 8.8, 9.9],
        None,
    ]
    assert ak.operations.to_list(ak.operations.flatten(cuda_array, axis=2)) == [
        [0.0, 1.1, None, 2.2, 3.3, 4.4],
        None,
        [],
        [5.5],
        None,
        [6.6, None, 7.7, 8.8, 9.9],
    ]
    assert ak.operations.to_list(ak.operations.flatten(cuda_array[1:], axis=2)) == [
        None,
        [],
        [5.5],
        None,
        [6.6, None, 7.7, 8.8, 9.9],
    ]
    assert ak.operations.to_list(ak.operations.flatten(cuda_array[:, 1:], axis=2)) == [
        [3.3, 4.4],
        None,
        [],
        [],
        None,
        [6.6, None, 7.7, 8.8, 9.9],
    ]

    content = ak.operations.from_iter(
        [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]], highlevel=False
    )
    index = ak.index.Index64(np.array([2, 1, 0, 3, 3, 4], dtype=np.int64))
    array = ak.contents.IndexedArray(index, content)

    cuda_array = ak.to_backend(array, "cuda")

    assert to_list(cuda_array) == [
        [3.3, 4.4],
        [],
        [0.0, 1.1, 2.2],
        [5.5],
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert ak.operations.to_list(ak.operations.flatten(cuda_array)) == [
        3.3,
        4.4,
        0.0,
        1.1,
        2.2,
        5.5,
        5.5,
        6.6,
        7.7,
        8.8,
        9.9,
    ]

    content = ak.operations.from_iter(
        [[[0.0, 1.1, 2.2], [], [3.3, 4.4]], [], [[5.5]], [[], [6.6, 7.7, 8.8, 9.9]]],
        highlevel=False,
    )
    index = ak.index.Index64(np.array([2, 2, 1, 0, 3], dtype=np.int64))
    array = ak.contents.IndexedArray(index, content)

    cuda_array = ak.to_backend(array, "cuda")

    assert to_list(cuda_array) == [
        [[5.5]],
        [[5.5]],
        [],
        [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
        [[], [6.6, 7.7, 8.8, 9.9]],
    ]
    assert ak.operations.to_list(ak.operations.flatten(cuda_array, axis=2)) == [
        [5.5],
        [5.5],
        [],
        [0.0, 1.1, 2.2, 3.3, 4.4],
        [6.6, 7.7, 8.8, 9.9],
    ]


def test_0150_flatten_RecordArray():
    array = ak.highlevel.Array(
        [
            {"x": [], "y": [[3, 3, 3]]},
            {"x": [[1]], "y": [[2, 2]]},
            {"x": [[2], [2]], "y": [[1]]},
            {"x": [[3], [3], [3]], "y": [[]]},
        ]
    ).layout

    cuda_array = ak.to_backend(array, "cuda")

    assert ak.operations.to_list(ak.operations.flatten(cuda_array, axis=2)) == [
        {"x": [], "y": [3, 3, 3]},
        {"x": [1], "y": [2, 2]},
        {"x": [2, 2], "y": [1]},
        {"x": [3, 3, 3], "y": []},
    ]
    assert ak.operations.to_list(ak.operations.flatten(cuda_array[1:], axis=2)) == [
        {"x": [1], "y": [2, 2]},
        {"x": [2, 2], "y": [1]},
        {"x": [3, 3, 3], "y": []},
    ]
    assert ak.operations.to_list(ak.operations.flatten(cuda_array[:, 1:], axis=2)) == [
        {"x": [], "y": []},
        {"x": [], "y": []},
        {"x": [2], "y": []},
        {"x": [3, 3], "y": []},
    ]



from __future__ import annotations

import pytest

import awkward as ak


def test_Array_mask_weakref():
    arr = ak.Array([1])
    m = arr.mask

    assert ak.all(m[[True]] == arr)

    del arr
    with pytest.raises(
        ValueError,
        match=r"The array to mask was deleted before it could be masked. If you want to construct this mask, you must either keep the array alive or use 'ak.mask' explicitly.",
    ):
        _ = m[[True]]
